/********************************************************************************
 * Copyright (c) 2004, Cha Zhang <czhang@ieee.org>, Advanced Multimedia 
 * Processing Lab, ECE, Carnegie Mellon University. 
 * All rights reserved. 
 * 
 * History
 * -------
 *
 * 2004.5.15 Created by Cha Zhang at CMU as a simplified version of the
 *           rendering algorithm implemented for the mobile camera array
 *           at AMP Lab, ECE, Carnegie Mellon University. 
 ********************************************************************************/

/********************************************************************************
 * This is the header file for basic image processing routines 
 ********************************************************************************/

#ifndef __CZ_IMAGE_UTILS_H__
#define __CZ_IMAGE_UTILS_H__

#include <stdlib.h>
#include <stdio.h>

#define CZ_IMAGE_MEM_ALLOCATED              1
#define CZ_IMAGE_MEM_NOT_ALLOCATED          0

typedef struct czsImageColorStruct
{   // color image 
    int w, h; 
    unsigned char *data;    // in the format of RGBRGBRGB...
} czsImgClr; 

typedef struct czsImageGrayStruct
{   // grayscale image 
    int w, h; 
    unsigned char *data; 
} czsImgGry; 

// create image space (memory), 
// return the pointer to the image, or 0 if allocation failed 
extern czsImgClr       *czfCreateImg(czsImgClr &img, int width, int height); 
extern czsImgGry       *czfCreateImg(czsImgGry &img, int width, int height); 

// release image space (memory)
extern void             czfReleaseImg(czsImgClr &img); 
extern void             czfReleaseImg(czsImgGry &img); 

/***************************************
 * PPM color image format read/write
 ***************************************/
// read the header of a PPM file, get the image width and height
// czfLoadImgPPM will call this function, so use it only if you just want to know the width and height of the image 
// return 0 if success, -1 if failed 
extern int              czfReadPPMHeader(FILE *fp, int &width, int &height); 
// load PPM image 
// return the pointer to the image, or 0 if loading failed 
extern czsImgClr       *czfLoadImgPPM(czsImgClr &img, const char *filename, int flag=CZ_IMAGE_MEM_NOT_ALLOCATED); 
// write PPM image 
// return the pointer to the image, or 0 if writing failed 
extern const czsImgClr *czfWriteImgPPM(const czsImgClr &img, const char *filename); 

/*******************************************
 * PGM grayscale image format read/write 
 *******************************************/
// read the header of a PGM file, get the image width and height
// czfLoadImgPGM will call this function, so use it only if you just want to know the width and height of the image 
// return 0 if success, -1 if failed 
extern int              czfReadPGMHeader(FILE *fp, int &width, int &height); 
// load PGM image 
// return the pointer to the image, or 0 if loading failed 
extern czsImgGry       *czfLoadImgPGM(czsImgGry &img, const char *filename, int flag=CZ_IMAGE_MEM_NOT_ALLOCATED); 
// write PGM image 
// return the pointer to the image, or 0 if writing failed 
extern const czsImgGry *czfWriteImgPGM(const czsImgGry &img, const char *filename); 

/**************************************************************************
 * BMP image format read/write 
 * NOTICE: these routines only load BMP files with 24-bit depth and 8-bit
 *         depth, in the 8-bit case, the pallete will be ingored and the 
 *         numbers read in will be considered directly as grayscale values 
 **************************************************************************/ 
// redefine basic structures only exist in Windows 
typedef struct czsBmpFileHeaderStruct
{
    unsigned short      bfType; 
    unsigned int        bfSize; 
    unsigned short      bfReserved1; 
    unsigned short      bfReserved2; 
    unsigned int        bfOffBits; 
} czsBmpFileHeader; 

typedef struct czsBmpInfoHeaderStruct
{
    unsigned int        biSize; 
    int                 biWidth; 
    int                 biHeight; 
    unsigned short      biPlanes; 
    unsigned short      biBitCount; 
    unsigned int        biCompression; 
    unsigned int        biSizeImage; 
    int                 biXPelsPerMeter; 
    int                 biYPelsPerMeter; 
    unsigned int        biClrUsed; 
    unsigned int        biClrImportant; 
} czsBmpInfoHeader; 

// read the header of a BMP file, get the width, height, and bit depth 
// czfLoadImgBMP will call this function
// return 0 if success, -1 if failed 
extern int              czfReadBMPHeader(FILE *fp, int &width, int &height, int &bpp); 
// load BMP color image
// return the pointer to the image, or 0 if loading failed 
extern czsImgClr       *czfLoadImgBMP(czsImgClr &img, const char *filename, int flag=CZ_IMAGE_MEM_NOT_ALLOCATED); 
// write BMP color image 
// NOTICE: writing BMP is SLOW due to RGB to BGR color space conversions and many file io operations!!! 
//         The latter could be improved at the cost of extra memories in a future implementation 
// return the pointer to the image, or 0 if writing failed 
extern const czsImgClr *czfWriteImgBMP(const czsImgClr &img, const char *filename); 
// load BMP grayscale image 
// return the pointer to the image, or 0 if loading failed 
extern czsImgGry       *czfLoadImgBMP(czsImgGry &img, const char *filename, int flag=CZ_IMAGE_MEM_NOT_ALLOCATED); 
// write BMP grayscale image 
// return the pointer to the image, or 0 if writing failed 
extern const czsImgGry *czfWriteImgBMP(const czsImgGry &img, const char *filename); 

/*******************************
 * Get pixel values 
 *******************************/
// retrieve color pixel value at an integer point 
// if w and h are invalid, clr[0]=clr[1]=clr[2]=-1 
// if you don't want to check w and h for speed concerns, write you own code to get the color 
// return the pointer of clr, or 0 if w and h are invalid
extern inline int      *czfGetPixel(const czsImgClr &img, int w, int h, int clr[3])
{
    if (w<0 || w>=img.w || h<0 || h>img.h)
    {
        clr[0]=clr[1]=clr[2]=-1; 
        return 0; 
    }
    unsigned char *p = &img.data[(h*img.w+w)*3]; 
    clr[0] = p[0], clr[1] = p[1], clr[2] = p[2]; 
    return clr; 
}; 
// retrieve color pixel value at arbitrary position using bilinear interpolation 
// if w and h are invalid, clr[0]=clr[1]=clr[2]=-1 
// return the pointer of clr, or 0 if w and h are invalid
extern double          *czfGetPixel(const czsImgClr &img, double w, double h, double clr[3]); 
// retrieve grayscale pixel value at an integer point 
// if w and h are invalid, gry=-1 
// return the pointer to gry, or 0 if w and h are invalid
extern inline int      *czfGetPixel(const czsImgGry &img, int w, int h, int &gry)
{
    if (w<0 || w>=img.w || h<0 || h>img.h)
    {
        gry = -1; 
        return 0; 
    }
    gry = img.data[h*img.w+w]; 
    return (&gry); 
}; 
// retrieve grayscale pixel value at arbitrary position using bilinear interpolation 
// if w and h are invalid, gry=-1 
// return the pointer to gry, or 0 if w and h are invalid
extern double          *czfGetPixel(const czsImgGry &img, double w, double h, double &gry); 

/******************************************
 * color converting macro and routines 
 ******************************************/
namespace cz_image_utils
{
    const int r_y = 0; 
    const int g_y = 256*1; 
    const int b_y = 256*2; 
    const int r_cb = 256*3; 
    const int g_cb = 256*4;
    const int b_cb = 256*5; 
    const int r_cr = b_cb; 
    const int g_cr = 256*6; 
    const int b_cr = 256*7; 

    const int cr_r = 0; 
    const int cb_g = 256*1; 
    const int cr_g = 256*2; 
    const int cb_b = 256*3; 

    static unsigned char fast_uchar[768]; 
    static int fast_rgb2ycc[256*8]; 
    static int fast_ycc2rgb[256*4]; 
    static int shift = 16; 
    static int init = 0; 
}

// always call the following function first before any color conversion 
// the cz_image_utils namespace defined above only has internal linkage for each cpp file, thus it is 
// necessary to initialize it in each cpp file if the following color conversions is ever used 
extern void             czfInitClrConvert(); 

#define CZ_RGB2YCBCR(src, dst)              \
{                                           \
    using namespace cz_image_utils;         \
    (dst)[0]=(fast_rgb2ycc[(src)[0]+r_y]+fast_rgb2ycc[(src)[1]+g_y]+fast_rgb2ycc[(src)[2]+b_y])>>shift;         \
    (dst)[1]=(fast_rgb2ycc[(src)[0]+r_cb]+fast_rgb2ycc[(src)[1]+g_cb]+fast_rgb2ycc[(src)[2]+b_cb])>>shift;      \
    (dst)[2]=(fast_rgb2ycc[(src)[0]+r_cr]+fast_rgb2ycc[(src)[1]+g_cr]+fast_rgb2ycc[(src)[2]+b_cr])>>shift;      \
}
#define CZ_RGB2GRY(src, dst)                \
{                                           \
    using namespace cz_image_utils;         \
    (dst)[0]=(fast_rgb2ycc[(src)[0]+r_y]+fast_rgb2ycc[(src)[1]+g_y]+fast_rgb2ycc[(src)[2]+b_y])>>shift;         \
}
#define CZ_YCBCR2RGB(src, dst)              \
{                                           \
    using namespace cz_image_utils;         \
    (dst)[0]=fast_uchar[(src)[0]+fast_ycc2rgb[(src)[2]+cr_r]+256];                                              \
    (dst)[1]=fast_uchar[(src)[0]+((fast_ycc2rgb[(src)[1]+cb_g]+fast_ycc2rgb[(src)[2]+cr_g])>>shift)+256];       \
    (dst)[2]=fast_uchar[(src)[0]+fast_ycc2rgb[(src)[1]+cb_b]+256];                                              \
}
#define CZ_YCBCR2RGB_INPLACE(buf)           \
{                                           \
    using namespace cz_image_utils;         \
    int y = (buf)[0], cb = (buf)[1];        \
    (buf)[0]=fast_uchar[y+fast_ycc2rgb[(buf)[2]+cr_r]+256];                                                     \
    (buf)[1]=fast_uchar[y+((fast_ycc2rgb[cb+cb_g]+fast_ycc2rgb[(buf)[2]+cr_g])>>shift)+256];                    \
    (buf)[2]=fast_uchar[y+fast_ycc2rgb[cb+cb_b]+256];                                                           \
}
#define CZ_YCBCR2GRY(src, dst)              \
{                                           \
    (dst)[0]=(src)[0];                      \
}
#define CZ_GRY2YCBCR(src, dst)              \
{                                           \
    (dst)[0]=(src)[0];                      \
    (dst)[1]=(dst)[2]=0;                    \
}
#define CZ_GRY2RGB(src, dst)                \
{                                           \
    (dst)[0]=(dst)[1]=(dst)[2]=(src)[0];    \
}

// convert a color image to grayscale image, this function assumes imgg has already been memory allocated
// return the pointer of imgg, or 0 if the two images' sizes do not match
extern czsImgGry       *czfClrImgToGry(const czsImgClr &imgc, czsImgGry &imgg); 
// convert a grayscale image to a color image 
extern czsImgClr       *czfGryImgToClr(const czsImgGry &imgg, czsImgClr &imgc); 

/**********************************************************
 * Routines to write/read little endian integers 
 **********************************************************/ 
// read a word 
extern inline unsigned short    czfReadWord(FILE *fp) 
{
    unsigned char b0 = getc(fp);
    unsigned char b1 = getc(fp);
    return ((b1 << 8) | b0);    
}
// read a dword
extern inline unsigned int      czfReadDWord(FILE *fp) 
{
    unsigned char b0 = getc(fp);
    unsigned char b1 = getc(fp);
    unsigned char b2 = getc(fp); 
    unsigned char b3 = getc(fp); 
    return ((((((b3 << 8) | b2) << 8) | b1) << 8) | b0);
}
// write a word, return 0 if success, -1 if fail
extern inline int               czfWriteWord(FILE *fp, unsigned short w)
{
    putc(w, fp); 
    return (putc(w >> 8, fp));
}
// write a dword, return 0 if success, -1 if fail 
extern inline int               czfWriteDWord(FILE *fp, unsigned int dw)
{
    putc(dw, fp); 
    putc(dw>>8, fp); 
    putc(dw>>16, fp); 
    return (putc(dw>>24, fp));
}

#endif 
