/********************************************************************************
 * Copyright (c) 2004, Cha Zhang <czhang@ieee.org>, Advanced Multimedia 
 * Processing Lab, ECE, Carnegie Mellon University. 
 * All rights reserved. 
 * 
 * History
 * -------
 *
 * 2004.5.15 Created by Cha Zhang at CMU as a simplified version of the
 *           rendering algorithm implemented for the mobile camera array
 *           at AMP Lab, ECE, Carnegie Mellon University. 
 ********************************************************************************/

/********************************************************************************
 * This is the header file for basic fixed size matrix operations etc. 
 ********************************************************************************/

#ifndef __CZ_MATH_UTILS_H__
#define __CZ_MATH_UTILS_H__

#include <math.h>
#include <stdlib.h>
#include <stdio.h>

/**************************
 * useful constants 
 **************************/
#define CZ_PI		        3.14159265358979	    // the variable pi
#define CZ_PIM2             6.28318530717959	    // 2 * pi 
#define CZ_PID2	            1.57079632679490	    // pi / 2 
#define CZ_E		        2.71828182845905	    // the variable e
#define CZ_SQRT2	        1.41421356237310	    // sqrt(2)
#define CZ_SQRT3	        1.73205080756888	    // sqrt(3) 
#define CZ_GOLDEN	        1.61803398874989	    // the golden ratio 
#define CZ_D2R	            0.01745329251994	    // convert degrees to radians 
#define CZ_R2D	            57.29577951308232	    // convert radians to degrees 
#define CZ_SMALL_NUMBER     1e-10                   // number that can be considered small
#define CZ_TINY_NUMBER      1e-20                   // number that is tiny

/************************** 
 * useful macros 
 **************************/
// absolute value of a
#define CZ_ABS(a)		(((a)<0) ? -(a) : (a))
// find minimum of a and b 
#define CZ_MIN2(a,b)	(((a)<(b))?(a):(b))	
// find maximum of a and b 
#define CZ_MAX2(a,b)	(((a)>(b))?(a):(b))	
// round a to nearest int 
#define CZ_ROUND(a)	((a)>0 ? (int)((a)+0.5) : -(int)(0.5-(a)))
// take sign of a, either -1, 0, or 1 
#define CZ_ZSGN(a)		(((a)<0) ? -1 : (a)>0 ? 1 : 0)	
// take binary sign of a, either -1, or 1 if >= 0 
#define CZ_SGN(a)		(((a)<0) ? -1 : 1)
// square a 
#define CZ_SQR(a)		((a)*(a))	
// swap a and b with intermediate variable t 
#define CZ_SWAP(a,b,t)  { (t)=(a); (a)=(b); (b)=(t); }
// linear interpolation from l (when a=0) to h (when a=1) (equal to (a*h)+((1-a)*l))
#define CZ_LERP(a,l,h)	((l)+(((h)-(l))*(a)))
// clamp the input to the specified range
#define CZ_CLAMP(v,l,h)	((v)<(l) ? (l) : (v) > (h) ? (h) : (v))
// find minimum of a, b and c
#define CZ_MIN3(a,b,c)  CZ_MIN2(CZ_MIN2(a,b),c) 
// find maximum of a, b and c
#define CZ_MAX3(a,b,c)  CZ_MAX2(CZ_MAX2(a,b),c)

/************************************
 * fixed size vector/matrix operation 
 ************************************/

/************************************
 * 2D/3D vector/matrix definition
 ************************************/
typedef struct czsPoint2Struct 
{	// 2d point/vector
    double x, y;
} czsPt2;
typedef czsPt2 czsVec2;

typedef struct czsPoint2HStruct
{   // 2d homogenious coordinate point 
    double x, y, w; 
} czsPt2H; 
typedef czsPt2H czsVec2H; 

// matrix elements are counted rowwise first in this library (different from OpenGL) 
typedef struct czsMatrix2Struct
{   // 2-by-2 matrix
    double m[4]; 
} czsMat2; 

typedef struct czsMatrix3Struct 
{	// 3-by-3 matrix
    double m[9];
} czsMat3;

typedef struct czsMatrix2x3Struct 
{   // 2-by-3 matrix 
    double m[6]; 
} czsMat2x3; 

/***********************************
 * 2D point/vector operations
 ***********************************/ 
// set a vector to zero 
extern inline void      czfV2SetZero(czsVec2 &v) { v.x=v.y=0.0; }; 
extern inline void      czfV2HSetZero(czsVec2H &v) { v.x=v.y=0.0, v.w=1.0; }; 
// set a vector to some value 
extern inline void      czfV2Set(czsVec2 &v, double x, double y) { v.x=x,v.y=y; }; 
extern inline void      czfV2HSet(czsVec2H &v, double x, double y) { v.x=x,v.y=y,v.w=1.0; }; 
extern inline void      czfV2Set(czsVec2 &v, const czsVec2 &a) { v.x=a.x,v.y=a.y; }; 
extern inline void      czfV2HSet(czsVec2H &v, const czsVec2H &a) { v.x=a.x,v.y=a.y,v.w=a.w; }; 
// negativate a vector 
extern inline void      czfV2Neg(czsVec2 &v) { v.x=-v.x,v.y=-v.y; }; 
// add two vectors 
extern inline void      czfV2Add(const czsVec2 &a, const czsVec2 &b, czsVec2 &v) { v.x=a.x+b.x,v.y=a.y+b.y; }; 
extern inline void      czfV2Add(czsVec2 &v, const czsVec2 &a) { v.x+=a.x,v.y+=a.y; };
// subtract two vectors 
extern inline void      czfV2Sub(const czsVec2 &a, const czsVec2 &b, czsVec2 &v) { v.x=a.x-b.x,v.y=a.y-b.y; }; 
extern inline void      czfV2Sub(czsVec2 &v, const czsVec2 &a) { v.x-=a.x,v.y-=a.y; }; 
// center point of two points 
extern inline void      czfV2CtrPt(const czsPt2 &a, const czsPt2 &b, czsPt2 &v) { v.x=(a.x+b.x)/2.0,v.y=(a.y+b.y)/2.0; }; 
// squared length of a vector 
extern inline double    czfV2SqrLen(const czsVec2 &v) { return(CZ_SQR(v.x)+CZ_SQR(v.y)); }; 
// length of a vector 
extern inline double    czfV2Len(const czsVec2 &v) { return(sqrt(CZ_SQR(v.x)+CZ_SQR(v.y))); }; 
// normalize a vector to length 1.0
extern inline czsVec2  *czfV2Norm(czsVec2 &v) 
{ 
    double val = czfV2Len(v); 
    if (val != 0.0) 
    { 
        val = 1.0/val; 
        v.x *= val;  
        v.y *= val; 
        return (&v); 
    }
    return 0; 
}; 
// normalize a homogenious coordinate 
extern inline czsVec2H *czfV2HNorm(czsVec2H &v)
{
    if (v.w != 0.0) 
    {
        double val = 1.0/v.w; 
        v.x *= val; 
        v.y *= val; 
        v.w = 1.0; 
        return (&v); 
    }
    return 0; 
}; 
// scale a vector to a certain length newlen
extern inline czsVec2  *czfV2Scale(czsVec2 &v, double newlen)
{
    double val = czfV2Len(v);
    if (val != 0.0) 
    { 
        val = newlen/val; 
        v.x *= val;   
        v.y *= val; 
        return (&v); 
    }
    return 0; 
}; 
// multiply the vector with a scalar 
extern inline void      czfV2MulScalar(czsVec2 &v, double s) { v.x*=s,v.y*=s; }; 
// dot product of two vectors 
extern inline double    czfV2Dot(const czsVec2 &a, const czsVec2 &b) { return(a.x*b.x+a.y*b.y); }; 
// angle between two vectors 
extern inline double    czfV2Angle(const czsVec2 &a, const czsVec2 &b)
{
    double dotprod = czfV2Dot(a,b); 
    if (dotprod == 0.0) 
        return 0.0; 
    return (acos(dotprod/(czfV2Len(a)*czfV2Len(b)))); 
}; 
// linearly interpolate between vectors by an amount alpha and return the resulting vector. 
// When alpha=0, v=lo.  When alpha=1, v=hi. 
extern inline void      czfV2Lerp(const czsVec2 &lo, const czsVec2 &hi, double alpha, czsVec2 &v)
{
    v.x = CZ_LERP(alpha, lo.x, hi.x);
    v.y = CZ_LERP(alpha, lo.y, hi.y); 
}; 
// make a linear combination of two vectors and return the result. 
// v = (a * ascl) + (b * bscl) 
extern inline void      czfV2Comb(const czsVec2 &a, const czsVec2 &b, double ascl, double bscl, czsVec2 &v ) 
{
    v.x = ascl*a.x + bscl*b.x;
    v.y = ascl*a.y + bscl*b.y; 
}; 
// multiply two vectors together component-wise 
extern inline void      czfV2Mul (const czsVec2 &a, const czsVec2 &b, czsVec2 &v) { v.x=a.x*b.x,v.y=a.y*b.y; }; 
// distance between two points 
extern inline double    czfV2Dist2Pts(const czsPt2 &a, const czsPt2 &b) { return(sqrt(CZ_SQR(a.x-b.x)+CZ_SQR(a.y-b.y))); };
// obtain the vector perpendicular to the input vector a 
extern inline void      czfV2Perp(const czsVec2 &a, czsVec2 &v) { v.x=-a.y,v.y=a.x; }; 

/****************************************
 * 2D vector and matrix multiplication 
 ****************************************/
// multiply a point by a 2x2 matrix and return the transformed point 
extern inline void      czfV2MulPtMat(czsPt2 &p, const czsMat2 &m)
{
    double val = p.x; 
    p.x = val*m.m[0] + p.y*m.m[1]; 
    p.y = val*m.m[2] + p.y*m.m[3]; 
}; 
// multiply a point by a 3x3 matrix and return the transformed point 
extern inline czsPt2   *czfV2MulPtMat(czsPt2 &p, const czsMat3 &m)
{
    double w = p.x*m.m[6] + p.y*m.m[7] + m.m[8]; 
    if (w != 0.0) 
    {
        w = 1.0/w; 
        double val = p.x; 
        p.x = (val*m.m[0] + p.y*m.m[1] + m.m[2]) * w;  
        p.y = (val*m.m[3] + p.y*m.m[4] + m.m[5]) * w; 
        return (&p); 
    }
    return 0; 
}; 
// multiply a homogenious coordinate point by a 3x3 matrix
extern inline void      czfV2HMulPtMat(czsPt2H &p, const czsMat3 &m)
{
    double val1 = p.x; 
    double val2 = p.y; 
    p.x = val1*m.m[0] + val2*m.m[1] + p.w*m.m[2]; 
    p.y = val1*m.m[3] + val2*m.m[4] + p.w*m.m[5];
    p.w = val1*m.m[6] + val2*m.m[7] + p.w*m.m[8]; 
}; 
// multiply a point by a 2x3 matrix and return the transformed point 
extern inline void      czfV2MulPtMat(czsPt2 &p, const czsMat2x3 &m)
{
    double val = p.x; 
    p.x = val*m.m[0] + p.y*m.m[1] + m.m[2]; 
    p.y = val*m.m[3] + p.y*m.m[4] + m.m[5]; 
}; 
// multiply a homogenious coordinate point by a 2x3 matrix 
extern inline void      czfV2MulPtMat(czsPt2H &p, const czsMat2x3 &m)
{
    double val = p.x;
    p.x = val*m.m[0] + p.y*m.m[1] + p.w*m.m[2];
    p.y = val*m.m[3] + p.y*m.m[4] + p.w*m.m[5]; 
    p.w = 1.0; 
}; 

/***********************************
 * 2x2 matrix operations
 ***********************************/ 
// find the determinant of a 2x2 matrix
extern inline double    czfDet2x2(double a, double b, double c, double d) { return(a*d-b*c); }; 
// find the determinant of a 3x3 matrix
extern inline double    czfDet3x3( double a1, double a2, double a3, 
                                   double b1, double b2, double b3, 
                                   double c1, double c2, double c3)
{
    return ( a1 * czfDet2x2( b2, b3, c2, c3 ) - 
             b1 * czfDet2x2( a2, a3, c2, c3 ) + 
             c1 * czfDet2x2( a2, a3, b2, b3 ) ); 
}; 
// set a 2x2 matrix to be zeros
extern inline void      czfM2SetZero(czsMat2 &m) { m.m[0]=m.m[1]=m.m[2]=m.m[3]=0.0; }; 
// set a 2x2 matrix to be identity
extern inline void      czfM2Identity(czsMat2 &m) { m.m[0]=m.m[3]=1.0, m.m[1]=m.m[2]=0.0; }; 
// set 2x2 matrix to some values
extern inline void      czfM2Set(czsMat2 &m, double a, double b, double c, double d)
{
    m.m[0]=a;
    m.m[1]=b;
    m.m[2]=c;
    m.m[3]=d; 
}; 
extern inline void      czfM2Set(czsMat2 &m, const czsMat2 &n) 
{ 
    m.m[0]=n.m[0]; 
    m.m[1]=n.m[1]; 
    m.m[2]=n.m[2];
    m.m[3]=n.m[3]; 
}; 
// transpose a 2x2 matrix
extern inline void      czfM2Transpose(czsMat2 &m) { double t; CZ_SWAP(m.m[1],m.m[2],t); };
extern inline void      czfM2Transpose(const czsMat2 &in, czsMat2 &out)
{
    out.m[0]=in.m[0]; 
    out.m[1]=in.m[2]; 
    out.m[2]=in.m[1]; 
    out.m[3]=in.m[3]; 
}; 
// get the adjoint matrix of a 2x2 matrix
extern inline void      czfM2Adjoint(const czsMat2 &in, czsMat2 &out)
{
    out.m[0]=in.m[3]; 
    out.m[1]=-in.m[1]; 
    out.m[2]=-in.m[2]; 
    out.m[3]=in.m[0]; 
}; 
// get the determinant of a 2x2 matrix 
extern inline double    czfM2Det(const czsMat2 &m) { return(czfDet2x2(m.m[0],m.m[1],m.m[2],m.m[3])); }; 
// inverse of a 2x2 matrix
extern czsMat2         *czfM2Inv(const czsMat2 &in, czsMat2 &out); 
// multiply two matrix together and get another, c=a*b 
extern inline void      czfM2MatMul(const czsMat2 &a, const czsMat2 &b, czsMat2 &c)
{
    c.m[0] = a.m[0]*b.m[0]+a.m[1]*b.m[2]; 
    c.m[1] = a.m[0]*b.m[1]+a.m[1]*b.m[3]; 
    c.m[2] = a.m[2]*b.m[0]+a.m[3]*b.m[2]; 
    c.m[3] = a.m[2]*b.m[1]+a.m[3]*b.m[3]; 
};

/***********************************
 * 3x3 matrix operations
 ***********************************/ 
// set a 3x3 matrix to be zeros
extern inline void      czfM3SetZero(czsMat3 &m)
{
    m.m[0]=m.m[1]=m.m[2]=m.m[3]=m.m[4]=m.m[5]=m.m[6]=m.m[7]=m.m[8]=0.0; 
}; 
// set a 3x3 matrix to identity matrix
extern inline void      czfM3Identity(czsMat3 &m)
{
    m.m[0]=m.m[4]=m.m[8]=1.0; 
    m.m[1]=m.m[2]=m.m[3]=m.m[5]=m.m[6]=m.m[7]=0.0; 
}; 
// set 3x3 matrix m to some value 
extern inline void      czfM3Set(czsMat3 &m, double a1, double a2, double a3,
                                             double b1, double b2, double b3,
                                             double c1, double c2, double c3)
{
    m.m[0] = a1, m.m[1] = a2, m.m[2] = a3; 
    m.m[3] = b1, m.m[4] = b2, m.m[5] = b3; 
    m.m[6] = c1, m.m[7] = c2, m.m[8] = c3; 
}; 
extern inline void      czfM3Set(czsMat3 &m, const czsMat3 &n) 
{
    m.m[0]=n.m[0], m.m[1]=n.m[1], m.m[2]=n.m[2]; 
    m.m[3]=n.m[3], m.m[4]=n.m[4], m.m[5]=n.m[5]; 
    m.m[6]=n.m[6], m.m[7]=n.m[7], m.m[8]=n.m[8]; 
}; 
// transpose a 3x3 matrix 
extern inline void      czfM3Transpose(czsMat3 &m) 
{
    double t; 
    CZ_SWAP(m.m[1],m.m[3],t);
    CZ_SWAP(m.m[2],m.m[4],t); 
    CZ_SWAP(m.m[5],m.m[7],t); 
};         
extern inline void      czfM3Transpose(const czsMat3 &in, czsMat3 &out)
{
    out.m[0]=in.m[0], out.m[1]=in.m[3], out.m[2]=in.m[6]; 
    out.m[3]=in.m[1], out.m[4]=in.m[4], out.m[5]=in.m[7]; 
    out.m[6]=in.m[2], out.m[7]=in.m[5], out.m[8]=in.m[8]; 
};
// get the adjoint matrix of a 3x3 matrix
extern void             czfM3Adjoint(const czsMat3 &in, czsMat3 &out); 
// get the determinant of a 3x3 matrix
extern inline double    czfM3Det(const czsMat3 &m)
{
    return (czfDet3x3(m.m[0],m.m[1],m.m[2],
                      m.m[3],m.m[4],m.m[5],
                      m.m[6],m.m[7],m.m[8])); 
}; 
// inverse a 3x3 matrix
extern czsMat3         *czfM3Inv(const czsMat3 &in, czsMat3 &out); 
// multiply matrices together c = a*b 
// note that c must point to NEITHER of the input matrices
extern void             czfM3MatMul(const czsMat3 &a, const czsMat3 &b, czsMat3 &c); 
extern void             czfM3MatMul(const czsMat2x3 &a, const czsMat3 &b, czsMat2x3 &c); 
extern void             czfM3MatMul(const czsMat2 &a, const czsMat2x3 &b, czsMat2x3 &c); 
// rotate a matrix by angle a (considered in 2D homogenious coord.), THIS IS PRE-CASCATENATE! 
extern void             czfM3RotPre(czsMat3 &m, double a); 
// rotate a matrix by angle a (considered in 2D homogenious coord.), THIS IS POST-CASCATENATE! 
extern void             czfM3RotPost(czsMat3 &m, double a); 
// scale a matrix by sx and sy, THIS IS PRE-CASCATENATE! 
extern void             czfM3ScalePre(czsMat3 &m, double sx, double sy); 
// scale a matrix by sx and sy, THIS IS POST-CASCATENATE! 
extern void             czfM3ScalePost(czsMat3 &m, double sx, double sy); 
// translate a matrix by tx, ty, THIS IS PRE-CASCATENATE! 
extern void             czM3TlPre(czsMat3 &m, double tx, double ty); 
// translate a matrix by tx, ty, THIS IS POST-CASCATENATE! 
extern void             czM3TlPost(czsMat3 &m, double tx, double ty); 

/************************************
 * 3d vector/matrix operation 
 ************************************/
typedef struct czsPoint3Struct 
{	// 3d point 
    double x, y, z;
} czsPt3;
typedef czsPt3 czsVec3;

typedef struct czsPoint3HStruct
{   // 3d point in homogenious coordinate 
    double x, y, z, w; 
} czsPt3H; 
typedef czsPt3H czsVec3H; 

typedef struct czsMatrix4Struct 
{	// 4-by-4 matrix 
    double m[16]; 
} czsMat4; 

typedef struct czsMatrix3x4Struct 
{   // 3-by-4 matrix 
    double m[12]; 
} czsMat3x4; 

/***********************************
 * 3D point/vector operations
 ***********************************/ 
// set the vector to be zero 
extern inline void      czfV3SetZero(czsVec3 &v) { v.x=v.y=v.z=0.0;  }; 
extern inline void      czfV3HSetZero(czsVec3H &v) { v.x=v.y=v.z=0.0; v.w=1.0; }; 
// set the vector value 
extern inline void      czfV3Set(czsVec3 &v, double x, double y, double z) { v.x=x,v.y=y,v.z=z; }; 
extern inline void      czfV3HSet(czsVec3H &v, double x, double y, double z) { v.x=x,v.y=y,v.z=z,v.w=1.0; }; 
extern inline void      czfV3Set(czsVec3 &v, const czsVec3 &a) { v.x=a.x,v.y=a.y,v.z=a.z; }; 
extern inline void      czfV3HSet(czsVec3H &v, const czsVec3H &a) { v.x=a.x,v.y=a.y,v.z=a.z,v.w=a.w; }; 
// negativate a vector 
extern inline void      czfV3Neg(czsVec3 &v) { v.x=-v.x,v.y=-v.y; }; 
// add two vectors 
extern inline void      czfV3Add(const czsVec3 &a, const czsVec3 &b, czsVec3 &v)
{
    v.x = a.x + b.x; 
    v.y = a.y + b.y; 
    v.z = a.z + b.z; 
};
extern inline void      czfV3Add(czsVec3 &v, const czsVec3 &a) { v.x+=a.x,v.y+=a.y,v.z+=a.z; };
// substract two vectors 
extern inline void      czfV3Sub(const czsVec3 &a, const czsVec3 &b, czsVec3 &v)
{
    v.x = a.x - b.x; 
    v.y = a.y - b.y; 
    v.z = a.z - b.z; 
}; 
extern inline void      czfV3Sub(czsVec3 &v, const czsVec3 &a) { v.x-=a.x,v.y-=a.y,v.z-=a.z; }; 
// center point of two points 
extern inline void      czfV3CtrPt(const czsPt3 &a, const czsPt3 &b, czsPt3 &v)
{
    v.x = (a.x+b.x)/2.0; 
    v.y = (a.y+b.y)/2.0; 
    v.z = (a.z+b.z)/2.0; 
}; 
// the squared length of a vector 
extern inline double    czfV3SqrLen(const czsVec3 &v) { return(CZ_SQR(v.x)+CZ_SQR(v.y)+CZ_SQR(v.z)); }; 
// length of a vector 
extern inline double    czfV3Len(const czsVec3 &v) { return(sqrt(CZ_SQR(v.x)+CZ_SQR(v.y)+CZ_SQR(v.z))); }; 
// normalizes a vector
extern inline czsVec3  *czfV3Norm(czsVec3 &v)
{
    double val = czfV3Len(v); 
    if (val != 0.0) 
    { 
        val = 1.0/val; 
        v.x *= val;  
        v.y *= val; 
        v.z *= val; 
        return (&v); 
    }
    return 0; 
}; 
// normalize a homogenious coordinate 
extern inline czsVec3H *czfV3HNorm(czsVec3H &v)
{
    if (v.w != 0.0) 
    {
        double val = 1.0/v.w; 
        v.x *= val; 
        v.y *= val; 
        v.z *= val; 
        v.w = 1.0; 
        return (&v); 
    }
    return 0; 
}; 
// scale a vector to a new length
extern inline czsVec3  *czfV3Scale(czsVec3 &v, double newlen)
{
    double val = czfV3Len(v);
    if (val != 0.0) 
    { 
        val = newlen/val; 
        v.x *= val;   
        v.y *= val; 
        v.z *= val; 
        return (&v); 
    }
    return 0; 
}; 
// multiple a vector with a scalar 
extern inline void      czfV3MulScalar (czsVec3 &v, double s) { v.x*=s,v.y*=s,v.z*=s; };
// dot product of two vectors 
extern inline double    czfV3Dot(const czsVec3 &a, const czsVec3 &b) { return(a.x*b.x+a.y*b.y+a.z*b.z); }; 
// angle between two vectors 
extern inline double    czfV3Angle(const czsVec3 &a, const czsVec3 &b) 
{
    double dotprod = czfV3Dot(a,b); 
    if (dotprod == 0.0) 
        return 0.0; 
    return (acos(dotprod/(czfV3Len(a)*czfV3Len(b)))); 
}
// linearly interpolate between vectors by an amount alpha
// When alpha=0, v=lo.  When alpha=1, v=hi. 
extern inline void      czfV3Lerp(const czsVec3 &lo, const czsVec3 &hi, double alpha, czsVec3 &v)
{
    v.x = CZ_LERP(alpha, lo.x, hi.x); 
    v.y = CZ_LERP(alpha, lo.y, hi.y); 
    v.z = CZ_LERP(alpha, lo.z, hi.z); 
}; 
// make a linear combination of two vectors, v = (a * ascl) + (b * bscl) 
extern inline void      czfV3Comb(const czsVec3 &a, const czsVec3 &b, double ascl, double bscl, czsVec3 &v)
{
    v.x = ascl*a.x + bscl*b.x;
    v.y = ascl*a.y + bscl*b.y;
    v.z = ascl*a.z + bscl*b.z;
}; 
// multiply two vectors together component-wise
extern inline void      czfV3Mul(const czsVec3 &a, const czsVec3 &b, czsVec3 &v)
{
    v.x = a.x*b.x; 
    v.y = a.y*b.y; 
    v.z = a.z*b.z; 
}; 
// the distance between two points 
extern inline double    czfV3Dist2Pts(const czsPt3 &a, const czsPt3 &b) 
{ 
    return(sqrt(CZ_SQR(a.x-b.x)+CZ_SQR(a.y-b.y)+CZ_SQR(a.z-b.z)));  
}; 
// the cross product v = a cross b 
extern inline void      czfV3Cross(const czsVec3 &a, const czsVec3 &b, czsVec3 &v)
{
    v.x = a.y*b.z - a.z*b.y; 
    v.y = a.z*b.x - a.x*b.z; 
    v.z = a.x*b.y - a.y*b.x; 
}; 

/****************************************
 * 3D vector and matrix multiplication 
 ****************************************/
// multiply a point by a 3x3 matrix 
extern inline void      czfV3MulPtMat(czsPt3 &p, const czsMat3 &m)
{
    double val1 = p.x;
    double val2 = p.y; 
    p.x = val1*m.m[0] + val2*m.m[1] + p.z*m.m[2]; 
    p.y = val1*m.m[3] + val2*m.m[4] + p.z*m.m[5]; 
    p.z = val1*m.m[6] + val2*m.m[7] + p.z*m.m[8]; 
}; 
// multiply a point by a 4x4 matrix (homogenious coord.) 
extern inline czsPt3   *czfV3MulPtMat(czsPt3 &p, const czsMat4 &m)
{
    double w = p.x*m.m[12] + p.y*m.m[13] + p.z*m.m[14] + m.m[15]; 
    if (w != 0.0) 
    {
        w = 1.0/w; 
        double val1 = p.x; 
        double val2 = p.y; 
        p.x = (val1*m.m[0] + val2*m.m[1] + p.z*m.m[2] + m.m[3]) * w;  
        p.y = (val1*m.m[4] + val2*m.m[5] + p.z*m.m[6] + m.m[7]) * w; 
        p.z = (val1*m.m[8] + val2*m.m[9] + p.z*m.m[10] + m.m[11]) * w; 
        return (&p); 
    }
    return 0; 
}; 
// multiply a homogenious coordinate point by a 4x4 matrix
extern inline void      czfV3HMulPtMat(czsPt3H &p, const czsMat4 &m)
{
    double val1 = p.x; 
    double val2 = p.y; 
    double val3 = p.z; 
    p.x = val1*m.m[0] + val2*m.m[1] + val3*m.m[2] + p.w*m.m[3]; 
    p.y = val1*m.m[4] + val2*m.m[5] + val3*m.m[6] + p.w*m.m[7]; 
    p.z = val1*m.m[8] + val2*m.m[9] + val3*m.m[10] + p.w*m.m[11]; 
    p.w = val1*m.m[12] + val2*m.m[13] + val3*m.m[14] + p.w*m.m[15]; 
}; 
// multiply a point by a 3x4 matrix 
extern inline void      czfV3MulPtMat(czsPt3 &p, const czsMat3x4 &m)
{
    double val1 = p.x;
    double val2 = p.y; 
    p.x = val1*m.m[0] + val2*m.m[1] + p.z*m.m[2] + m.m[3]; 
    p.y = val1*m.m[4] + val2*m.m[5] + p.z*m.m[6] + m.m[7]; 
    p.z = val1*m.m[8] + val2*m.m[9] + p.z*m.m[10] + m.m[11]; 
}; 
// multiply a homogenious coordinate point by a 3x4 matrix 
extern inline void      czfV3MulPtMat(czsPt3H &p, const czsMat3x4 &m)
{
    double val1 = p.x;
    double val2 = p.y; 
    p.x = val1*m.m[0] + val2*m.m[1] + p.z*m.m[2] + p.w*m.m[3]; 
    p.y = val1*m.m[4] + val2*m.m[5] + p.z*m.m[6] + p.w*m.m[7]; 
    p.z = val1*m.m[8] + val2*m.m[9] + p.z*m.m[10] + p.w*m.m[11]; 
    p.w = 1.0; 
}; 
// project a 3D point to a 2D position 
extern inline czsPt2   *czfV3Proj2V2(const czsPt3 &p3, const czsMat3x4 &m, czsPt2 &p2) 
{
    double w = p3.x*m.m[8] + p3.y*m.m[9] + p3.z*m.m[10] + m.m[11]; 
    if (w != 0.0) 
    {
        w = 1.0/w; 
        p2.x = (p3.x*m.m[0] + p3.y*m.m[1] + p3.z*m.m[2] + m.m[3]) * w;  
        p2.y = (p3.x*m.m[4] + p3.y*m.m[5] + p3.z*m.m[6] + m.m[7]) * w; 
        return (&p2); 
    }
    czfV2SetZero(p2); 
    return 0; 
}
// project a 3D homogenious coordinate point to a 2D position 
extern inline czsPt2   *czfV3Proj2V2(const czsPt3H &p3, const czsMat3x4 &m, czsPt2 &p2) 
{
    double w = p3.x*m.m[8] + p3.y*m.m[9] + p3.z*m.m[10] + p3.w*m.m[11]; 
    if (w != 0.0) 
    {
        w = 1.0/w; 
        p2.x = (p3.x*m.m[0] + p3.y*m.m[1] + p3.z*m.m[2] + p3.w*m.m[3]) * w;  
        p2.y = (p3.x*m.m[4] + p3.y*m.m[5] + p3.z*m.m[6] + p3.w*m.m[7]) * w; 
        return (&p2); 
    }
    czfV2SetZero(p2); 
    return 0; 
}

/***********************************
 * 4x4 matrix operations
 ***********************************/ 
// set 4x4 matrix to be zero matrix
extern inline void      czfM4SetZero(czsMat4 &m)
{
    m.m[0] = m.m[1] = m.m[2] = m.m[3] = 
        m.m[4] = m.m[5] = m.m[6] = m.m[7] = 
        m.m[8] = m.m[9] = m.m[10] = m.m[11] = 
        m.m[12] = m.m[13] = m.m[14] = m.m[14] = 0.0; 
}; 
// set 4x4 matrix to be identity matrix
extern inline void      czfM4Identity(czsMat4 &m)
{
    m.m[0] = m.m[5] = m.m[10] = m.m[15] = 1.0; 
    m.m[1] = m.m[2] = m.m[3] = m.m[4] = m.m[6] = m.m[7] = 
        m.m[8] = m.m[9] = m.m[11] = m.m[12] = m.m[13] = m.m[14] = 0.0; 
}; 
// set matrix m to some value 
extern inline void      czfM4Set(czsMat4 &m, double a1, double a2, double a3, double a4, 
                                             double b1, double b2, double b3, double b4, 
                                             double c1, double c2, double c3, double c4, 
                                             double d1, double d2, double d3, double d4)
{
    m.m[0] = a1, m.m[1] = a2, m.m[2] = a3, m.m[3] = a4; 
    m.m[4] = b1, m.m[5] = b2, m.m[6] = b3, m.m[7] = b4; 
    m.m[8] = c1, m.m[9] = c2, m.m[10] = c3, m.m[11] = c4; 
    m.m[12] = d1, m.m[13] = d2, m.m[14] = d3, m.m[15] = d4; 
}; 
extern inline void      czfM4Set(czsMat4 &m, const czsMat4 &n)
{
    m.m[0]=n.m[0], m.m[1]=n.m[1], m.m[2]=n.m[2], m.m[3]=n.m[3]; 
    m.m[4]=n.m[4], m.m[5]=n.m[5], m.m[6]=n.m[6], m.m[7]=n.m[7]; 
    m.m[8]=n.m[8], m.m[9]=n.m[9], m.m[10]=n.m[10], m.m[11]=n.m[11]; 
    m.m[12]=n.m[12], m.m[13]=n.m[13], m.m[14]=n.m[14], m.m[15]=n.m[15]; 
}; 
extern inline void      czfM3x4Set(czsMat3x4 &m, double a1, double a2, double a3, double a4, 
                                                 double b1, double b2, double b3, double b4, 
                                                 double c1, double c2, double c3, double c4)
{
    m.m[0] = a1, m.m[1] = a2, m.m[2] = a3, m.m[3] = a4; 
    m.m[4] = b1, m.m[5] = b2, m.m[6] = b3, m.m[7] = b4; 
    m.m[8] = c1, m.m[9] = c2, m.m[10] = c3, m.m[11] = c4; 
}; 
extern inline void      czfM3x4Set(czsMat3x4 &m, const czsMat3x4 &n)
{
    m.m[0]=n.m[0], m.m[1]=n.m[1], m.m[2]=n.m[2], m.m[3]=n.m[3]; 
    m.m[4]=n.m[4], m.m[5]=n.m[5], m.m[6]=n.m[6], m.m[7]=n.m[7]; 
    m.m[8]=n.m[8], m.m[9]=n.m[9], m.m[10]=n.m[10], m.m[11]=n.m[11]; 
}
// transpose a 4x4 matrix 
extern inline void      czfM4Transpose(czsMat4 &m) 
{
    double t; 
    CZ_SWAP(m.m[1],m.m[4],t); 
    CZ_SWAP(m.m[2],m.m[8],t); 
    CZ_SWAP(m.m[3],m.m[12],t); 
    CZ_SWAP(m.m[6],m.m[9],t); 
    CZ_SWAP(m.m[7],m.m[13],t); 
    CZ_SWAP(m.m[11],m.m[14],t); 
}; 
extern inline void      czfM4Transpose(const czsMat4 &in, czsMat4 &out)
{
    out.m[0]=in.m[0], out.m[1]=in.m[4], out.m[2]=in.m[8], out.m[3]=in.m[12]; 
    out.m[4]=in.m[1], out.m[5]=in.m[5], out.m[6]=in.m[9], out.m[7]=in.m[13]; 
    out.m[8]=in.m[2], out.m[9]=in.m[6], out.m[10]=in.m[10], out.m[11]=in.m[14]; 
    out.m[12]=in.m[3], out.m[13]=in.m[7], out.m[14]=in.m[11], out.m[15]=in.m[15]; 
}; 
// find the adjoint matrix of a 4x4 matrix 
extern void             czfM4Adjoint(const czsMat4 &in, czsMat4 &out); 
// get the determinant of a 4x4 matrix
extern double           czfM4Det(const czsMat4 &m); 
// inverse a 4x4 matrix
extern czsMat4         *czfM4Inv(const czsMat4 &in, czsMat4 &out); 
// multiply together matrices c = a*b, note that c must be different from a and b 
extern void             czfM4MatMul(const czsMat4 &a, const czsMat4 &b, czsMat4 &c); 
extern void             czfM4MatMul(const czsMat3x4 &a, const czsMat4 &b, czsMat3x4 &c); 
extern void             czfM4MatMul(const czsMat3 &a, const czsMat3x4 &b, czsMat3x4 &c); 
// rotate 4x4 matrix along x-axis by angle a, THIS IS PRE-CASCATENATE! 
extern void             czfM4RotXPre(czsMat4 &m, double a); 
// rotate 4x4 matrix along x-axis by angle a, THIS IS POST-CASCATENATE! 
extern void             czfM4RotXPost(czsMat4 &m, double a); 
// rotate 4x4 matrix along y-axis by angle a, THIS IS PRE-CASCATENATE! 
extern void             czfM4RotYPre(czsMat4 &m, double a);
// rotate 4x4 matrix along y-axis by angle a, THIS IS POST-CASCATENATE! 
extern void             czfM4RotYPost(czsMat4 &m, double a);
// rotate 4x4 matrix along z-axis by angle a, THIS IS PRE-CASCATENATE! 
extern void             czfM4RotZPre(czsMat4 &m, double a);
// rotate 4x4 matrix along z-axis by angle a, THIS IS POST-CASCATENATE! 
extern void             czfM4RotZPost(czsMat4 &m, double a);
// scale a 4x4 matrix by sx, sy and sz, THIS IS PRE-CASCATENATE! 
extern void             czfM4ScalePre (czsMat4 &m, double sx, double sy, double sz); 
// scale a 4x4 matrix by sx, sy and sz, THIS IS POST-CASCATENATE! 
extern void             czfM4ScalePost (czsMat4 &m, double sx, double sy, double sz); 
// translate a 4x4 matrix by tx, ty and tz, THIS IS PRE-CASCATENATE! 
extern void             czfM4TlPre(czsMat4 &m, double tx, double ty, double tz); 
// translate a 4x4 matrix by tx, ty and tz, THIS IS POST-CASCATENATE! 
extern void             czfM4TlPost(czsMat4 &m, double tx, double ty, double tz); 

/*********************************
 * quaternion operation 
 *********************************/
typedef struct czsQuaternionStruct
{   /* quaternion */
    double x, y, z, w; 
} czsQuat; 

// set a quaternion
extern inline void      czfQuatSet(czsQuat &q, double x, double y, double z, double w)
{
    q.x=x, q.y=y, q.z=z, q.w=w; 
}; 
extern inline void      czfQuatSet(czsQuat &q, const czsQuat &a)
{
    q.x=a.x, q.y=a.y, q.z=a.z, q.w=a.w; 
}; 
// normalize a quaternion 
extern inline czsQuat  *czfQuatNorm(czsQuat &q)
{
    double val = sqrt(CZ_SQR(q.x)+CZ_SQR(q.y)+CZ_SQR(q.z)+CZ_SQR(q.w)); 
    if (val != 0.0) 
    { 
        val = 1.0/val; 
        q.x *= val;  
        q.y *= val; 
        q.z *= val;  
        q.w *= val; 
        return (&q); 
    }
    return(0);
};
// compose two quaternions
extern inline czsQuat  *czfQuatComp(const czsQuat &q1, const czsQuat &q2, czsQuat &q3)
{
    q3.w = (q1.w*q2.w) - (q1.x*q2.x) - (q1.y*q2.y) - (q1.z*q2.z); 
    q3.x = (q1.w*q2.x) + (q1.x*q2.w) + (q1.y*q2.z) - (q1.z*q2.y); 
    q3.y = (q1.w*q2.y) - (q1.x*q2.z) + (q1.y*q2.w) + (q1.z*q2.x); 
    q3.z = (q1.w*q2.z) + (q1.x*q2.y) - (q1.y*q2.x) + (q1.z*q2.w); 
	if (czfQuatNorm(q3))
        return (&q3); 
    else
        return 0; 
}; 
// rotate vector v by quaternion q
extern czsVec3         *czfQuatRot(czsVec3 &v, const czsQuat &q); 
// convert a quaternion to a rotation matrix 
extern void             czfQuatToRotMat(czsQuat &q, czsMat3 &R); 
extern void             czfRotMatToQuat(czsMat3 &R, czsQuat &q); 
extern void             czfQuatToAngAxis(czsQuat &q, czsVec3 &v); 
extern void             czfAngAxisToQuat(czsVec3 &v, czsQuat &q); 

/***********************************
 * the Rodrigues formula 
 ***********************************/ 
// convert from rotation angle and axis to a rotation matrix, and reverse, these functions use the Rodrigues formula
extern void             czfAngAxisToRotMat(czsVec3 &vec, czsMat3 &R); 
extern void             czfRotMatToAngAxis(czsMat3 &R, czsVec3 &v);

/***********************************
 * miscellaneous small functions 
 ***********************************/
extern inline int        czfRoundUpPow2(int x) 
{
    if (x <= 0) return 0; 
    int i = 1; 
    while (i < x) 
        i = (i << 1); 
    return i; 
}; 

extern inline int        czfRoundUpLog2(int x) 
{
    if (x <= 0) return -1; 
    int i=1, k=0; 
    while (i < x) 
    {
        i = (i << 1); 
        k++; 
    }
    return k; 
}; 

#endif
