/********************************************************************************
 * Copyright (c) 2004, Cha Zhang <czhang@ieee.org>, Advanced Multimedia 
 * Processing Lab, ECE, Carnegie Mellon University. 
 * All rights reserved. 
 * 
 * History
 * -------
 *
 * 2004.5.15 Created by Cha Zhang at CMU as a simplified version of the
 *           rendering algorithm implemented for the mobile camera array
 *           at AMP Lab, ECE, Carnegie Mellon University. 
 ********************************************************************************/

/********************************************************************************
 * This is the cpp file for basic fixed size matrix operations etc. 
 ********************************************************************************/

#include "cz_math_utils.h" 

// inverse of a 2x2 matrix
czsMat2 *czfM2Inv(const czsMat2 &in, czsMat2 &out) 
{
    double det = czfM2Det(in); 
    if ( fabs(det) < CZ_TINY_NUMBER) 
    {
        printf("Matrix un-invertable!\n"); 
        czfM2SetZero(out); 
        return 0; 
    }
    det = 1.0/det; 
    out.m[0] =  in.m[3]*det; 
    out.m[1] = -in.m[1]*det; 
    out.m[2] = -in.m[2]*det; 
    out.m[3] =  in.m[0]*det; 
    return (&out); 
}

// get the adjoint matrix of a 3x3 matrix
void czfM3Adjoint(const czsMat3 &in, czsMat3 &out)
{
    const double *pin = in.m; 
    double *pout = out.m; 
    pout[0] =  czfDet2x2(pin[4],pin[5],pin[7],pin[8]); 
    pout[1] = -czfDet2x2(pin[1],pin[2],pin[7],pin[8]); 
    pout[2] =  czfDet2x2(pin[1],pin[2],pin[4],pin[5]); 
    pout[3] = -czfDet2x2(pin[3],pin[5],pin[6],pin[8]); 
    pout[4] =  czfDet2x2(pin[0],pin[2],pin[6],pin[8]); 
    pout[5] = -czfDet2x2(pin[0],pin[2],pin[3],pin[5]);         
    pout[6] =  czfDet2x2(pin[3],pin[4],pin[6],pin[7]); 
    pout[7] = -czfDet2x2(pin[0],pin[1],pin[6],pin[7]); 
    pout[8] =  czfDet2x2(pin[0],pin[1],pin[3],pin[4]); 
}

// inverse a 3x3 matrix
czsMat3 *czfM3Inv(const czsMat3 &in, czsMat3 &out)
{
    double det = czfM3Det(in); 
    if ( fabs(det) < CZ_TINY_NUMBER) 
    {
        printf ("Matrix un-invertable!\n"); 
        czfM3SetZero(out); 
        return 0; 
    }
    det = 1.0/det; 
    // find the adjoint matrix
    czfM3Adjoint(in, out); 

    // scale the adjoint matrix by the determinant
    for (int i=0; i<9; i++)
        out.m[i] *= det; 
    return (&out); 
}

// multiply together matrices c = ab 
// note that c must point to NEITHER of the input matrices
void czfM3MatMul(const czsMat3 &a, const czsMat3 &b, czsMat3 &c)
{
    const double *pa = a.m;
    const double *pb = b.m;
    double *pc = c.m; 
    pc[0] = pa[0]*pb[0] + pa[1]*pb[3] + pa[2]*pb[6]; 
    pc[1] = pa[0]*pb[1] + pa[1]*pb[4] + pa[2]*pb[7]; 
    pc[2] = pa[0]*pb[2] + pa[1]*pb[5] + pa[2]*pb[8]; 
    pc[3] = pa[3]*pb[0] + pa[4]*pb[3] + pa[5]*pb[6]; 
    pc[4] = pa[3]*pb[1] + pa[4]*pb[4] + pa[5]*pb[7]; 
    pc[5] = pa[3]*pb[2] + pa[4]*pb[5] + pa[5]*pb[8]; 
    pc[6] = pa[6]*pb[0] + pa[7]*pb[3] + pa[8]*pb[6]; 
    pc[7] = pa[6]*pb[1] + pa[7]*pb[4] + pa[8]*pb[7]; 
    pc[8] = pa[6]*pb[2] + pa[7]*pb[5] + pa[8]*pb[8]; 
}

void czfM3MatMul(const czsMat2x3 &a, const czsMat3 &b, czsMat2x3 &c)
{
    const double *pa = a.m;
    const double *pb = b.m;
    double *pc = c.m; 
    pc[0] = pa[0]*pb[0] + pa[1]*pb[3] + pa[2]*pb[6]; 
    pc[1] = pa[0]*pb[1] + pa[1]*pb[4] + pa[2]*pb[7]; 
    pc[2] = pa[0]*pb[2] + pa[1]*pb[5] + pa[2]*pb[8]; 
    pc[3] = pa[3]*pb[0] + pa[4]*pb[3] + pa[5]*pb[6]; 
    pc[4] = pa[3]*pb[1] + pa[4]*pb[4] + pa[5]*pb[7]; 
    pc[5] = pa[3]*pb[2] + pa[4]*pb[5] + pa[5]*pb[8]; 
}

void czfM3MatMul(const czsMat2 &a, const czsMat2x3 &b, czsMat2x3 &c)
{
    const double *pa = a.m;
    const double *pb = b.m;
    double *pc = c.m; 
    pc[0] = pa[0]*pb[0] + pa[1]*pb[3]; 
    pc[1] = pa[0]*pb[1] + pa[1]*pb[4];
    pc[2] = pa[0]*pb[2] + pa[1]*pb[5]; 
    pc[3] = pa[2]*pb[0] + pa[3]*pb[3]; 
    pc[4] = pa[2]*pb[1] + pa[3]*pb[4];
    pc[5] = pa[2]*pb[2] + pa[3]*pb[5]; 
}

// rotate a matrix by angle a (considered in 2D homogenious coord.), THIS IS PRE-CASCATENATE! 
void czfM3RotPre(czsMat3 &m, double a)
{
    double c = cos (a);
    double s = sin (a); 
    double val = m.m[0]; 
    m.m[0] = val*c - m.m[3]*s; 
    m.m[3] = val*s + m.m[3]*c; 
    val = m.m[1]; 
    m.m[1] = val*c - m.m[4]*s; 
    m.m[4] = val*s + m.m[4]*c; 
    val = m.m[2]; 
    m.m[2] = val*c - m.m[5]*s; 
    m.m[5] = val*s + m.m[5]*c; 
}

// rotate a matrix by angle a (considered in 2D homogenious coord.), THIS IS POST-CASCATENATE! 
void czfM3RotPost(czsMat3 &m, double a)
{
    double c = cos (a);
    double s = sin (a); 
    double val = m.m[0]; 
    m.m[0] = val*c + m.m[1]*s; 
    m.m[1] = -val*s + m.m[1]*c; 
    val = m.m[3]; 
    m.m[3] = val*c + m.m[4]*s; 
    m.m[4] = -val*s + m.m[4]*c; 
    val = m.m[6]; 
    m.m[6] = val*c + m.m[7]*s; 
    m.m[7] = -val*s + m.m[7]*c; 
}

// scale a matrix by sx and sy, THIS IS PRE-CASCATENATE! 
void czfM3ScalePre(czsMat3 &m, double sx, double sy)
{
    m.m[0] *= sx; 
    m.m[1] *= sx; 
    m.m[2] *= sx; 
    m.m[3] *= sy; 
    m.m[4] *= sy; 
    m.m[5] *= sy; 
}

// scale a matrix by sx and sy, THIS IS POST-CASCATENATE! 
void czfM3ScalePost(czsMat3 &m, double sx, double sy)
{
    m.m[0] *= sx; 
    m.m[3] *= sx; 
    m.m[6] *= sx; 
    m.m[1] *= sy; 
    m.m[4] *= sy; 
    m.m[7] *= sy; 
}

// translate a matrix by tx, ty, THIS IS PRE-CASCATENATE! 
void czM3TlPre(czsMat3 &m, double tx, double ty)
{
    m.m[0] += m.m[6]*tx; 
    m.m[1] += m.m[7]*tx; 
    m.m[2] += m.m[8]*tx; 
    m.m[3] += m.m[6]*ty; 
    m.m[4] += m.m[7]*ty; 
    m.m[5] += m.m[8]*ty; 
}	

// translate a matrix by tx, ty, THIS IS POST-CASCATENATE! 
void czM3TlPost(czsMat3 &m, double tx, double ty)
{
    m.m[2] += m.m[0]*tx + m.m[1]*ty; 
    m.m[5] += m.m[3]*tx + m.m[4]*ty; 
    m.m[8] += m.m[6]*tx + m.m[7]*ty; 
}	

// find the adjoint matrix of a 4x4 matrix 
void czfM4Adjoint(const czsMat4 &in, czsMat4 &out)
{    
    const double *pin = in.m; 
    double *pout = out.m; 
    pout[0]  =   czfDet3x3(pin[5],pin[6],pin[7],pin[9],pin[10],pin[11],pin[13],pin[14],pin[15]); 
    pout[1]  = - czfDet3x3(pin[1],pin[2],pin[3],pin[9],pin[10],pin[11],pin[13],pin[14],pin[15]); 
    pout[2]  =   czfDet3x3(pin[1],pin[2],pin[3],pin[5],pin[6],pin[7],pin[13],pin[14],pin[15]); 
    pout[3]  = - czfDet3x3(pin[1],pin[2],pin[3],pin[5],pin[6],pin[7],pin[9],pin[10],pin[11]); 
        
    pout[4]  = - czfDet3x3(pin[4],pin[6],pin[7],pin[8],pin[10],pin[11],pin[12],pin[14],pin[15]); 
    pout[5]  =   czfDet3x3(pin[0],pin[2],pin[3],pin[8],pin[10],pin[11],pin[12],pin[14],pin[15]); 
    pout[6]  = - czfDet3x3(pin[0],pin[2],pin[3],pin[4],pin[6],pin[7],pin[12],pin[14],pin[15]); 
    pout[7]  =   czfDet3x3(pin[0],pin[2],pin[3],pin[4],pin[6],pin[7],pin[8],pin[10],pin[11]); 
        
    pout[8]  =   czfDet3x3(pin[4],pin[5],pin[7],pin[8],pin[9],pin[11],pin[12],pin[13],pin[15]); 
    pout[9]  = - czfDet3x3(pin[0],pin[1],pin[3],pin[8],pin[9],pin[11],pin[12],pin[13],pin[15]);
    pout[10] =   czfDet3x3(pin[0],pin[1],pin[3],pin[4],pin[5],pin[7],pin[12],pin[13],pin[15]);
    pout[11] = - czfDet3x3(pin[0],pin[1],pin[3],pin[4],pin[5],pin[7],pin[8],pin[9],pin[11]);
        
    pout[12] = - czfDet3x3(pin[4],pin[5],pin[6],pin[8],pin[9],pin[10],pin[12],pin[13],pin[14]); 
    pout[13] =   czfDet3x3(pin[0],pin[1],pin[2],pin[8],pin[9],pin[10],pin[12],pin[13],pin[14]); 
    pout[14] = - czfDet3x3(pin[0],pin[1],pin[2],pin[4],pin[5],pin[6],pin[12],pin[13],pin[14]); 
    pout[15] =   czfDet3x3(pin[0],pin[1],pin[2],pin[4],pin[5],pin[6],pin[8],pin[9],pin[10]); 
}

// get the determinant of a 4x4 matrix
double czfM4Det(const czsMat4 &m)
{
    const double *p = m.m; 
    double ans = p[0] * czfDet3x3(p[5],p[6],p[7],p[9],p[10],p[11],p[13],p[14],p[15])
        - p[4] * czfDet3x3(p[1],p[2],p[3],p[9],p[10],p[11],p[13],p[14],p[15]) 
        + p[8] * czfDet3x3(p[1],p[2],p[3],p[5],p[6],p[7],p[13],p[14],p[15]) 
        - p[12] * czfDet3x3(p[1],p[2],p[3],p[5],p[6],p[7],p[9],p[10],p[11]); 
    return ans;
}

// inverse a 4x4 matrix
czsMat4 *czfM4Inv(const czsMat4 &in, czsMat4 &out)
{
    double det = czfM4Det(in); 
    if ( fabs(det) < CZ_TINY_NUMBER) 
    {
        printf ("Matrix un-invertable!\n"); 
        czfM4SetZero(out); 
        return 0; 
    }
    det = 1.0/det; 
    // find the adjoint matrix
    czfM4Adjoint(in, out); 

    // scale the adjoint matrix by the determinant
    for (int i=0; i<16; i++)
        out.m[i] *= det; 
    return (&out); 
}

// multiply together matrices c = a*b, note that c must be different from a and b 
void czfM4MatMul(const czsMat4 &a, const czsMat4 &b, czsMat4 &c)
{
    const double *pa = a.m;
    const double *pb = b.m;
    double *pc = c.m; 
    pc[0]  = pa[0]*pb[0] + pa[1]*pb[4] + pa[2]*pb[8] + pa[3]*pb[12]; 
    pc[1]  = pa[0]*pb[1] + pa[1]*pb[5] + pa[2]*pb[9] + pa[3]*pb[13]; 
    pc[2]  = pa[0]*pb[2] + pa[1]*pb[6] + pa[2]*pb[10] + pa[3]*pb[14]; 
    pc[3]  = pa[0]*pb[3] + pa[1]*pb[7] + pa[2]*pb[11] + pa[3]*pb[15]; 
    pc[4]  = pa[4]*pb[0] + pa[5]*pb[4] + pa[6]*pb[8] + pa[7]*pb[12]; 
    pc[5]  = pa[4]*pb[1] + pa[5]*pb[5] + pa[6]*pb[9] + pa[7]*pb[13]; 
    pc[6]  = pa[4]*pb[2] + pa[5]*pb[6] + pa[6]*pb[10] + pa[7]*pb[14]; 
    pc[7]  = pa[4]*pb[3] + pa[5]*pb[7] + pa[6]*pb[11] + pa[7]*pb[15]; 
    pc[8]  = pa[8]*pb[0] + pa[9]*pb[4] + pa[10]*pb[8] + pa[11]*pb[12]; 
    pc[9]  = pa[8]*pb[1] + pa[9]*pb[5] + pa[10]*pb[9] + pa[11]*pb[13]; 
    pc[10] = pa[8]*pb[2] + pa[9]*pb[6] + pa[10]*pb[10] + pa[11]*pb[14]; 
    pc[11] = pa[8]*pb[3] + pa[9]*pb[7] + pa[10]*pb[11] + pa[11]*pb[15]; 
    pc[12] = pa[12]*pb[0] + pa[13]*pb[4] + pa[14]*pb[8] + pa[15]*pb[12]; 
    pc[13] = pa[12]*pb[1] + pa[13]*pb[5] + pa[14]*pb[9] + pa[15]*pb[13]; 
    pc[14] = pa[12]*pb[2] + pa[13]*pb[6] + pa[14]*pb[10] + pa[15]*pb[14]; 
    pc[15] = pa[12]*pb[3] + pa[13]*pb[7] + pa[14]*pb[11] + pa[15]*pb[15]; 
}

// multiply together matrices c = a*b 
void czfM4MatMul(const czsMat3x4 &a, const czsMat4 &b, czsMat3x4 &c)
{
    const double *pa = a.m;
    const double *pb = b.m;
    double *pc = c.m; 
    pc[0]  = pa[0]*pb[0] + pa[1]*pb[4] + pa[2]*pb[8] + pa[3]*pb[12]; 
    pc[1]  = pa[0]*pb[1] + pa[1]*pb[5] + pa[2]*pb[9] + pa[3]*pb[13]; 
    pc[2]  = pa[0]*pb[2] + pa[1]*pb[6] + pa[2]*pb[10] + pa[3]*pb[14]; 
    pc[3]  = pa[0]*pb[3] + pa[1]*pb[7] + pa[2]*pb[11] + pa[3]*pb[15]; 
    pc[4]  = pa[4]*pb[0] + pa[5]*pb[4] + pa[6]*pb[8] + pa[7]*pb[12]; 
    pc[5]  = pa[4]*pb[1] + pa[5]*pb[5] + pa[6]*pb[9] + pa[7]*pb[13]; 
    pc[6]  = pa[4]*pb[2] + pa[5]*pb[6] + pa[6]*pb[10] + pa[7]*pb[14]; 
    pc[7]  = pa[4]*pb[3] + pa[5]*pb[7] + pa[6]*pb[11] + pa[7]*pb[15]; 
    pc[8]  = pa[8]*pb[0] + pa[9]*pb[4] + pa[10]*pb[8] + pa[11]*pb[12]; 
    pc[9]  = pa[8]*pb[1] + pa[9]*pb[5] + pa[10]*pb[9] + pa[11]*pb[13]; 
    pc[10] = pa[8]*pb[2] + pa[9]*pb[6] + pa[10]*pb[10] + pa[11]*pb[14]; 
    pc[11] = pa[8]*pb[3] + pa[9]*pb[7] + pa[10]*pb[11] + pa[11]*pb[15]; 
}

// multiply together matrices c = a*b 
void czfM4MatMul(const czsMat3 &a, const czsMat3x4 &b, czsMat3x4 &c)
{
    const double *pa = a.m;
    const double *pb = b.m;
    double *pc = c.m; 
    pc[0]  = pa[0]*pb[0] + pa[1]*pb[4] + pa[2]*pb[8]; 
    pc[1]  = pa[0]*pb[1] + pa[1]*pb[5] + pa[2]*pb[9]; 
    pc[2]  = pa[0]*pb[2] + pa[1]*pb[6] + pa[2]*pb[10];
    pc[3]  = pa[0]*pb[3] + pa[1]*pb[7] + pa[2]*pb[11]; 
    pc[4]  = pa[3]*pb[0] + pa[4]*pb[4] + pa[5]*pb[8]; 
    pc[5]  = pa[3]*pb[1] + pa[4]*pb[5] + pa[5]*pb[9]; 
    pc[6]  = pa[3]*pb[2] + pa[4]*pb[6] + pa[5]*pb[10];
    pc[7]  = pa[3]*pb[3] + pa[4]*pb[7] + pa[5]*pb[11]; 
    pc[8]  = pa[6]*pb[0] + pa[7]*pb[4] + pa[8]*pb[8]; 
    pc[9]  = pa[6]*pb[1] + pa[7]*pb[5] + pa[8]*pb[9]; 
    pc[10] = pa[6]*pb[2] + pa[7]*pb[6] + pa[8]*pb[10];
    pc[11] = pa[6]*pb[3] + pa[7]*pb[7] + pa[8]*pb[11]; 
}

// rotate 4x4 matrix along x-axis by angle a, THIS IS PRE-CASCATENATE! 
void czfM4RotXPre(czsMat4 &m, double a)
{
    double c = cos (a);
    double s = sin (a); 
    double val = m.m[4]; 
    m.m[4] = val*c - m.m[8]*s; 
    m.m[8] = val*s + m.m[8]*c; 
    val = m.m[5]; 
    m.m[5] = val*c - m.m[9]*s; 
    m.m[9] = val*s + m.m[9]*c; 
    val = m.m[6]; 
    m.m[6] = val*c - m.m[10]*s; 
    m.m[10] = val*s + m.m[10]*c; 
    val = m.m[7]; 
    m.m[7] = val*c - m.m[11]*s; 
    m.m[11] = val*s + m.m[11]*c; 
}

// rotate 4x4 matrix along x-axis by angle a, THIS IS POST-CASCATENATE! 
void czfM4RotXPost(czsMat4 &m, double a)
{
    double c = cos (a);
    double s = sin (a); 
    double val = m.m[1]; 
    m.m[1] = val*c + m.m[2]*s; 
    m.m[2] = -val*s + m.m[2]*c; 
    val = m.m[5]; 
    m.m[5] = val*c + m.m[6]*s; 
    m.m[6] = -val*s + m.m[6]*c; 
    val = m.m[9]; 
    m.m[9] = val*c + m.m[10]*s; 
    m.m[10] = -val*s + m.m[10]*c; 
    val = m.m[13]; 
    m.m[13] = val*c + m.m[14]*s; 
    m.m[14] = -val*s + m.m[14]*c; 
}

// rotate 4x4 matrix along y-axis by angle a, THIS IS PRE-CASCATENATE! 
void czfM4RotYPre(czsMat4 &m, double a) 
{
    double c = cos (a);
    double s = sin (a); 
    double val = m.m[0]; 
    m.m[0] = val*c + m.m[8]*s; 
    m.m[8] = -val*s + m.m[8]*c; 
    val = m.m[1]; 
    m.m[1] = val*c + m.m[9]*s; 
    m.m[9] = -val*s + m.m[9]*c; 
    val = m.m[2]; 
    m.m[2] = val*c + m.m[10]*s; 
    m.m[10] = -val*s + m.m[10]*c; 
    val = m.m[3]; 
    m.m[3] = val*c + m.m[11]*s; 
    m.m[11] = -val*s + m.m[11]*c; 
}

// rotate 4x4 matrix along y-axis by angle a, THIS IS POST-CASCATENATE! 
void czfM4RotYPost(czsMat4 &m, double a) 
{
    double c = cos (a);
    double s = sin (a); 
    double val = m.m[0]; 
    m.m[0] = val*c - m.m[2]*s; 
    m.m[2] = val*s + m.m[2]*c; 
    val = m.m[4]; 
    m.m[4] = val*c - m.m[6]*s; 
    m.m[6] = val*s + m.m[6]*c; 
    val = m.m[8]; 
    m.m[8] = val*c - m.m[10]*s; 
    m.m[10] = val*s + m.m[10]*c; 
    val = m.m[12]; 
    m.m[12] = val*c - m.m[14]*s; 
    m.m[14] = val*s + m.m[14]*c; 
}

// rotate 4x4 matrix along z-axis by angle a, THIS IS PRE-CASCATENATE! 
void czfM4RotZPre(czsMat4 &m, double a) 
{
    double c = cos (a);
    double s = sin (a); 
    double val = m.m[0]; 
    m.m[0] = val*c - m.m[4]*s; 
    m.m[4] = val*s + m.m[4]*c; 
    val = m.m[1]; 
    m.m[1] = val*c - m.m[5]*s; 
    m.m[5] = val*s + m.m[5]*c; 
    val = m.m[2]; 
    m.m[2] = val*c - m.m[6]*s; 
    m.m[6] = val*s + m.m[6]*c; 
    val = m.m[3]; 
    m.m[3] = val*c - m.m[7]*s; 
    m.m[7] = val*s + m.m[7]*c; 
}

// rotate 4x4 matrix along z-axis by angle a, THIS IS POST-CASCATENATE! 
void czfM4RotZPost(czsMat4 &m, double a) 
{
    double c = cos (a);
    double s = sin (a); 
    double val = m.m[0]; 
    m.m[0] = val*c + m.m[1]*s; 
    m.m[1] = -val*s + m.m[1]*c; 
    val = m.m[4]; 
    m.m[4] = val*c + m.m[5]*s; 
    m.m[5] = -val*s + m.m[5]*c; 
    val = m.m[8]; 
    m.m[8] = val*c + m.m[9]*s; 
    m.m[9] = -val*s + m.m[9]*c; 
    val = m.m[12]; 
    m.m[12] = val*c + m.m[13]*s; 
    m.m[13] = -val*s + m.m[13]*c; 
}

// scale a 4x4 matrix by sx, sy and sz, THIS IS PRE-CASCATENATE! 
void czfM4ScalePre (czsMat4 &m, double sx, double sy, double sz)
{
    m.m[0]*=sx, m.m[1]*=sx, m.m[2]*=sx, m.m[3]*=sx; 
    m.m[4]*=sy, m.m[5]*=sy, m.m[6]*=sy, m.m[7]*=sy; 
    m.m[8]*=sz, m.m[9]*=sz, m.m[10]*=sz, m.m[11]*=sz; 
}

// scale a 4x4 matrix by sx, sy and sz, THIS IS POST-CASCATENATE! 
void czfM4ScalePost (czsMat4 &m, double sx, double sy, double sz)
{
    m.m[0]*=sx, m.m[4]*=sx, m.m[8]*=sx, m.m[12]*=sx; 
    m.m[1]*=sy, m.m[5]*=sy, m.m[9]*=sy, m.m[13]*=sy; 
    m.m[2]*=sz, m.m[6]*=sz, m.m[10]*=sz, m.m[14]*=sz; 
}

// translate a 4x4 matrix by tx, ty and tz, THIS IS PRE-CASCATENATE! 
void czfM4TlPre(czsMat4 &m, double tx, double ty, double tz)
{
    m.m[0]+=m.m[12]*tx, m.m[1]+=m.m[13]*tx, m.m[2]+=m.m[14]*tx, m.m[3]+=m.m[15]*tx; 
    m.m[4]+=m.m[12]*ty, m.m[5]+=m.m[13]*ty, m.m[6]+=m.m[14]*ty, m.m[7]+=m.m[15]*ty; 
    m.m[8]+=m.m[12]*tz, m.m[9]+=m.m[13]*tz, m.m[10]+=m.m[14]*tz, m.m[11]+=m.m[15]*tz; 
}

// translate a 4x4 matrix by tx, ty and tz, THIS IS POST-CASCATENATE! 
void czfM4TlPost(czsMat4 &m, double tx, double ty, double tz)
{
    m.m[3] += m.m[0]*tx + m.m[1]*ty + m.m[2]*tz; 
    m.m[7] += m.m[4]*tx + m.m[5]*ty + m.m[6]*tz; 
    m.m[11] += m.m[8]*tx + m.m[9]*ty + m.m[10]*tz; 
    m.m[15] += m.m[12]*tx + m.m[13]*ty + m.m[14]*tz; 
}

// rotate vector v by quaternion q
czsVec3 *czfQuatRot(czsVec3 &v, const czsQuat &q)
{
    double len = czfV3Len(v); 
    if (len == 0.0) 
    {
        printf ("Cannot rotate a vector of length 0!\n"); 
        return 0; 
    }
    czsQuat p, qi, qtmp; 
    czfQuatSet(p, v.x, v.y, v.z, 0); 
    czfQuatSet(qi, -q.x, -q.y, -q.z, q.w); 
    czfQuatComp (q, p, qtmp); 
    czfQuatComp (qtmp, qi, p); 
    v.x = p.x, v.y = p.y, v.z = p.z; 
    czfV3Scale(v, len); 
    return (&v); 
}

void czfQuatToRotMat(czsQuat &q, czsMat3 &R)
{
    double xx = q.x*q.x; 
    double xy = q.x*q.y; 
    double xz = q.x*q.z; 
    double xw = q.x*q.w; 
    double yy = q.y*q.y; 
    double yz = q.y*q.z; 
    double yw = q.y*q.w; 
    double zz = q.z*q.z; 
    double zw = q.z*q.w; 

    double *p = R.m; 
    p[0] = 1-2*(yy+zz); 
    p[1] = 2*(xy+zw); 
    p[2] = 2*(xz-yw); 
    p[3] = 2*(xy-zw); 
    p[4] = 1-2*(xx+zz); 
    p[5] = 2*(yz+xw); 
    p[6] = 2*(xz+yw); 
    p[7] = 2*(yz-xw); 
    p[8] = 1-2*(xx+yy); 
}

void czfRotMatToQuat(czsMat3 &R, czsQuat &q)
{
    double *p = R.m; 
    double t = 1 + p[0] + p[4] + p[8]; 

    if (t > CZ_SMALL_NUMBER) 
    {
        double s = sqrt(t)*2; 
        q.x = (p[5]-p[7])/s; 
        q.y = (p[6]-p[2])/s;
        q.z = (p[1]-p[3])/s; 
        q.w = 0.25*s; 
        return; 
    }

    if (p[0] > p[4] && p[0] > p[8])     // column 0 dominate 
    {
        double s = sqrt(1.0 + p[0] - p[4] - p[8]); 
        q.x = 0.25*s; 
        q.y = (p[1]+p[3])/s; 
        q.z = (p[6]+p[2])/s; 
        q.w = (p[5]-p[7])/s; 
    }
    else if (p[4] > p[8]) 
    {
        double s = sqrt(1.0 + p[4] - p[0] - p[8]); 
        q.x = (p[1]+p[3])/s; 
        q.y = 0.25*s; 
        q.z = (p[5]+p[7])/s; 
        q.w = (p[6]-p[2])/s; 
    }
    else 
    {
        double s = sqrt(1.0 + p[8] - p[0] - p[4]); 
        q.x = (p[6]+p[2])/s; 
        q.y = (p[5]+p[7])/s; 
        q.z = 0.25*s;   
        q.w = (p[1]-p[3])/s; 
    }
}

void czfQuatToAngAxis(czsQuat &q, czsVec3 &v)
{
    czfQuatNorm(q); 
    double cos_a = q.w; 
    double theta = acos(cos_a)*2; 
    double sin_a = sqrt(1-cos_a*cos_a); 
    if (sin_a < CZ_SMALL_NUMBER)
        sin_a = 1; 
    v.x = q.x*theta/sin_a; 
    v.y = q.y*theta/sin_a; 
    v.z = q.z*theta/sin_a; 
}

void czfAngAxisToQuat(czsVec3 &v, czsQuat &q) 
{
    double theta = czfV3Len(v); 
    czfV3Norm(v); 
    double sin_a = sin(theta/2); 
    double cos_a = cos(theta/2); 
    q.x = v.x*sin_a; 
    q.y = v.y*sin_a;
    q.z = v.z*sin_a;
    q.w = cos_a; 
}

void czfAngAxisToRotMat(czsVec3 &vec, czsMat3 &R)
{
    double theta = czfV3Len(vec); 
    if (theta  < CZ_SMALL_NUMBER) 
    {
        czfM3Identity(R); 
        return; 
    }
    double c = cos(theta); 
    double s = sin(theta); 
    double u = vec.x/theta; 
    double v = vec.y/theta; 
    double w = vec.z/theta; 
    double c1 = 1-c;
    double *p = R.m; 
    p[0] = u*u*c1 + c; 
    p[1] = u*v*c1 - w*s; 
    p[2] = u*w*c1 + v*s; 
    p[3] = u*v*c1 + w*s; 
    p[4] = v*v*c1 + c; 
    p[5] = v*w*c1 - u*s;
    p[6] = u*w*c1 - v*s; 
    p[7] = v*w*c1 + u*s; 
    p[8] = w*w*c1 + c; 
}

void czfRotMatToAngAxis(czsMat3 &R, czsVec3 &v)
{
    double *p = R.m; 
    double tr = (p[0]+p[4]+p[8]-1)/2; 
    double theta = acos(tr); 
    double s = sin(theta); 
    if (s > CZ_SMALL_NUMBER)
    {
        double vth = 1.0/(2*s); 
        v.x = (p[7]-p[5])*vth*theta; 
        v.y = (p[2]-p[6])*vth*theta; 
        v.z = (p[3]-p[1])*vth*theta; 
    }
    else
    {
        if (tr > 0) 
            v.x = v.y = v.z = 0.0; 
        else 
        {
            v.x = sqrt((p[0]+1)/2.0)*theta; 
            v.y = sqrt((p[4]+1)/2.0*(2*((p[1]>=0)? 1:0)-1))*theta; 
            v.z = sqrt((p[8]+1)/2.0*(2*((p[2]>=0)? 1:0)-1))*theta; 
        }
    }
}
