/********************************************************************************
 * Copyright (c) 2004, Cha Zhang <czhang@ieee.org>, Advanced Multimedia 
 * Processing Lab, ECE, Carnegie Mellon University. 
 * All rights reserved. 
 * 
 * History
 * -------
 *
 * 2004.5.15 Created by Cha Zhang at CMU as a simplified version of the
 *           rendering algorithm implemented for the mobile camera array
 *           at AMP Lab, ECE, Carnegie Mellon University. 
 ********************************************************************************/

/********************************************************************************
 * This is the cpp file for basic image processing routines 
 ********************************************************************************/

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "cz_math_utils.h"
#include "cz_image_utils.h" 

#define CZ_MAX_IMAGE_WIDTH      8192
#define CZ_MAX_IMAGE_HEIGHT     8192

// create image space (memory) for color image
czsImgClr *czfCreateImg(czsImgClr &img, int width, int height)
{
    if (width<1 || width>CZ_MAX_IMAGE_WIDTH || height<1 || height>CZ_MAX_IMAGE_HEIGHT) 
    {
        printf ("Image size out of range! Valid width 1~%d, height 1~%d.\n",CZ_MAX_IMAGE_WIDTH,CZ_MAX_IMAGE_HEIGHT); 
        return 0;       // don't create image that is too large 
    }

    if ((img.data = (unsigned char *)malloc(width*height*3)) != NULL) 
    {
        img.w = width; 
        img.h = height; 
        return (&img); 
    }
    printf ("Failed to allocate memory!\n"); 
    return 0; 
}

// create image space (memory) for grayscale image
czsImgGry *czfCreateImg(czsImgGry &img, int width, int height)
{
    if (width<1 || width>CZ_MAX_IMAGE_WIDTH || height<1 || height>CZ_MAX_IMAGE_HEIGHT) 
    {
        printf ("Image size out of range! Valid width 1~%d, height 1~%d.\n",CZ_MAX_IMAGE_WIDTH,CZ_MAX_IMAGE_HEIGHT); 
        return 0;       // don't create images that are too large 
    }

    if ((img.data = (unsigned char *)malloc(width*height)) != NULL)
    {
        img.w = width; 
        img.h = height; 
        return (&img); 
    }
    printf ("Failed to allocate memory!\n"); 
    return 0; 
}

// release image space (memory)
void czfReleaseImg(czsImgClr &img)
{
    if (img.data != NULL) 
        free (img.data); 
    img.data = NULL; 
    img.w = 0; 
    img.h = 0; 
}

void czfReleaseImg(czsImgGry &img)
{
    if (img.data != NULL) 
        free (img.data); 
    img.data = NULL; 
    img.w = 0; 
    img.h = 0; 
}

// read the header of a PPM file, get the image width and height 
int czfReadPPMHeader(FILE *fp, int &width, int &height)
{
    char ch;
    if (fscanf(fp, "P%c\n", &ch) != 1 || ch != '6') 
    {
        printf ("The image file is not in PPM raw format!\n"); 
        return -1; 
    }

    // skip all the comments 
    ch = getc(fp);
    while (ch == '#')
    {
        do {
            ch = getc(fp);
        } while (ch != '\n');	// read to the end of the line, safer than fgets 
        ch = getc(fp);
    }

    if (!isdigit(ch))
    {
        printf ("Unable to read PPM header information (width and height)!\n"); 
        return -1; 
    }

    ungetc(ch, fp);		// put that digit back 

    // read the width, height, and maximum value for a pixel 
    int  maxval;
    fscanf(fp, "%d%d%d\n", &width, &height, &maxval); 

    if (maxval != 255)
    {
        printf ("Unable to deal with PPM images that are not 24-bit true color!\n"); 
        return -1; 
    }
    return 0;   
}

// load PPM color image 
czsImgClr *czfLoadImgPPM(czsImgClr &img, const char *filename, int flag)
{
    FILE *fp; 
    if ((fp=fopen(filename, "rb")) == NULL)
    {
        printf ("Unable to open image file: %s!\n", filename); 
        return 0; 
    }
    int width, height; 
    if (czfReadPPMHeader(fp, width, height)) 
    {
        printf ("Unable to read the PPM image header for file: %s!\n", filename); 
        fclose (fp); 
        return 0; 
    }
    
    if (flag == CZ_IMAGE_MEM_NOT_ALLOCATED) 
    {
        if (!czfCreateImg(img, width, height)) 
        {
            fclose (fp); 
            return 0; 
        }
    }
    else 
    {
        if (img.w != width || img.h != height) 
        {
            printf ("The size of the image does not match with that of the image file %s!\n", filename); 
            fclose (fp); 
            return 0; 
        }
    }

    int size = width*height*3; 
    fseek (fp, -size, SEEK_END);    // this is because while reading the header, fscanf may read extra bytes 
    int readsize = fread((void *)img.data, 1, (size_t)size, fp); 

    if (readsize != size) 
    {
        printf ("Unable to read all the data from the image file %s!\n", filename); 
        if (flag == CZ_IMAGE_MEM_NOT_ALLOCATED) 
            czfReleaseImg(img); 
        fclose (fp); 
        return 0; 
    }
    fclose(fp);
    return (&img); 
}

// write PPM image 
const czsImgClr *czfWriteImgPPM(const czsImgClr &img, const char *filename) 
{
    FILE *fp; 
    if ((fp=fopen(filename, "wb")) == NULL)
    {
        printf ("Unable to open file %s for writing!\n", filename); 
        return 0; 
    }

    fprintf(fp, "P6\n%d %d\n%d\n", img.w, img.h, 255); 
    int size = img.w*img.h*3; 
    int writesize = fwrite((void *)img.data, 1, (size_t)size, fp); 

    if (writesize != size) 
    {
        printf ("Unable to write all the data to the image file %s!\n", filename); 
        fclose (fp); 
        return 0; 
    }  
    fclose(fp);
    return (&img); 
}

// read the header of a PGM file, get the image width and height 
int czfReadPGMHeader(FILE *fp, int &width, int &height)
{
    char ch;
    if (fscanf(fp, "P%c\n", &ch) != 1 || ch != '5') 
    {
        printf ("The image file is not in PGM raw format!\n"); 
        return -1; 
    }

    // skip all the comments 
    ch = getc(fp);
    while (ch == '#')
    {
        do {
            ch = getc(fp);
        } while (ch != '\n');	// read to the end of the line, safer than fgets 
        ch = getc(fp);
    }

    if (!isdigit(ch))
    {
        printf ("Unable to read PGM header information (width and height)!\n"); 
        return -1; 
    }

    ungetc(ch, fp);		// put that digit back 

    // read the width, height, and maximum value for a pixel 
    int  maxval;
    fscanf(fp, "%d%d%d\n", &width, &height, &maxval); 

    if (maxval != 255)
    {
        printf ("Unable to deal with PGM images that are not 8-bit grayscale!\n"); 
        return -1; 
    }
    return 0;   
}

// load PGM grayscale image 
czsImgGry *czfLoadImgPGM(czsImgGry &img, const char *filename, int flag)
{
    FILE *fp; 
    if ((fp=fopen(filename, "rb")) == NULL)
    {
        printf ("Unable to open image file: %s!\n", filename); 
        return 0; 
    }
    int width, height; 
    if (czfReadPGMHeader(fp, width, height)) 
    {
        printf ("Unable to read the PGM image header for file: %s!\n", filename); 
        fclose (fp); 
        return 0; 
    }
    
    if (flag == CZ_IMAGE_MEM_NOT_ALLOCATED) 
    {
        if (!czfCreateImg(img, width, height)) 
        {
            fclose (fp); 
            return 0; 
        }
    }
    else 
    {
        if (img.w != width || img.h != height) 
        {
            printf ("The size of the image does not match with that of the image file %s!\n", filename); 
            fclose (fp); 
            return 0; 
        }
    }

    int size = width*height; 
    fseek (fp, -size, SEEK_END);    // this is because while reading the header, fscanf may read extra bytes 
    int readsize = fread((void *)img.data, 1, (size_t)size, fp); 

    if (readsize != size) 
    {
        printf ("Unable to read all the data from the image file %s!\n", filename); 
        if (flag == CZ_IMAGE_MEM_NOT_ALLOCATED) 
            czfReleaseImg(img); 
        fclose (fp); 
        return 0; 
    }
    fclose(fp);
    return (&img); 
}

// write PGM image 
const czsImgGry *czfWriteImgPGM(const czsImgGry &img, const char *filename) 
{
    FILE *fp; 
    if ((fp=fopen(filename, "wb")) == NULL)
    {
        printf ("Unable to open file %s for writing!\n", filename); 
        return 0; 
    }

    fprintf(fp, "P5\n%d %d\n%d\n", img.w, img.h, 255); 
    int size = img.w*img.h; 
    int writesize = fwrite((void *)img.data, 1, (size_t)size, fp); 

    if (writesize != size) 
    {
        printf ("Unable to write all the data to the image file %s!\n", filename); 
        fclose (fp); 
        return 0; 
    }  
    fclose(fp);
    return (&img); 
}

// read the header of a BMP file, get the width, height, and bit depth 
int czfReadBMPHeader(FILE *fp, int &width, int &height, int &bpp)
{
    // there are padding problem and endian problem while reading the header
    // so we read the stuff in one by one 
    czsBmpFileHeader fh; 
    fh.bfType = czfReadWord(fp); 
    fh.bfSize = czfReadDWord(fp); 
    fh.bfReserved1 = czfReadWord(fp); 
    fh.bfReserved2 = czfReadWord(fp); 
    fh.bfOffBits = czfReadDWord(fp); 

    if (fh.bfType != 0x4d42 || fh.bfReserved1 != 0 || fh.bfReserved2 != 0) 
    {
        printf ("The file is not a valid BMP file!\n"); 
        return -1; 
    }
    
    czsBmpInfoHeader ih; 
    ih.biSize = czfReadDWord(fp); 
    ih.biWidth = (int)czfReadDWord(fp); 
    ih.biHeight = (int)czfReadDWord(fp); 
    ih.biPlanes = czfReadWord(fp); 
    ih.biBitCount = czfReadWord(fp); 
    ih.biCompression = czfReadDWord(fp); 
    ih.biSizeImage = czfReadDWord(fp); 
    ih.biXPelsPerMeter = (int)czfReadDWord(fp); 
    ih.biYPelsPerMeter = (int)czfReadDWord(fp); 
    ih.biClrUsed = czfReadDWord(fp); 
    ih.biClrImportant = czfReadDWord(fp); 
    if ((ih.biBitCount != 24) && (ih.biBitCount != 8)) 
    {
        printf ("Unsupported BMP bit depth! Currently we only handle 8 or 24 bit images!\n"); 
        return -1; 
    }
    if (ih.biCompression != 0)  // 0 corresponds to BI_RGB in Windows 
    {
        printf ("Unsupported BMP format! Currently we only read uncompressed images!\n"); 
        return -1; 
    }
    width = ih.biWidth; 
    height = ih.biHeight; 
    bpp = ih.biBitCount;    
    return 0; 
}

// load BMP color image
czsImgClr *czfLoadImgBMP(czsImgClr &img, const char *filename, int flag)
{
    FILE *fp; 
    if ((fp=fopen(filename, "rb")) == NULL)
    {
        printf ("Unable to open image file: %s!\n", filename); 
        return 0; 
    }
    int width, height, bpp; 
    if (czfReadBMPHeader(fp, width, height,bpp)) 
    {
        printf ("Unable to read the BMP image header for file: %s!\n", filename); 
        fclose (fp); 
        return 0; 
    }
    if (bpp != 24) 
    {
        printf ("The BMP image is not color!\n"); 
        fclose (fp);    
        return 0; 
    }
    if (flag == CZ_IMAGE_MEM_NOT_ALLOCATED) 
    {
        if (!czfCreateImg(img, width, height)) 
        {
            fclose (fp); 
            return 0; 
        }
    }
    else 
    {
        if (img.w != width || img.h != height) 
        {
            printf ("The size of the image does not match with that of the image file %s!\n", filename); 
            fclose (fp); 
            return 0; 
        }
    }

    // now read the data row by row, notice BMP stores pixels bottom up 
    for (int row=height-1; row>=0; row--) 
    {
        int size = width*3; 
        unsigned char *p = &img.data[row*size]; 
        int readsize = fread((void *)p, 1, (size_t)size, fp); 
        if (readsize != size) 
        {
            printf ("Error while reading image data! \n"); 
            if (flag == CZ_IMAGE_MEM_NOT_ALLOCATED) 
                czfReleaseImg(img); 
            fclose(fp); 
            return 0; 
        }
        // unfortunately, BMP is following BGRBGR... format and we have to swap it here 
        unsigned char t; 
        for (int i=0; i<width; i++, p+=3) 
            CZ_SWAP(p[0],p[2],t); 
        // BMP has padding for each row such that each row is 4-byte aligned
        unsigned char pad[4];   // no more than 4 bytes of padding 
        if (size%4) 
        {
            size = 4-size%4; 
            readsize = fread((void *)pad, 1, (size_t)size, fp); 
            if (readsize != size) 
            {
                printf ("Error while reading image data! \n"); 
                if (flag == CZ_IMAGE_MEM_NOT_ALLOCATED) 
                    czfReleaseImg(img); 
                fclose(fp); 
                return 0; 
            }
        }
    }    
    fclose(fp);
    return (&img); 
}

// write BMP color image 
const czsImgClr *czfWriteImgBMP(const czsImgClr &img, const char *filename)
{
    FILE *fp; 
    if ((fp=fopen(filename, "wb")) == NULL)
    {
        printf ("Unable to open file %s for writing!\n", filename); 
        return 0; 
    }
    
    int padwidth = (img.w*24+31)/32*4;  // width*3 padded to 4 bytes 
    czsBmpFileHeader fh; 
    fh.bfType = 0x4d42; 
    fh.bfSize = 14+40+padwidth*img.h; 
    fh.bfReserved1 = 0; 
    fh.bfReserved2 = 0; 
    fh.bfOffBits = 14+40; 
    czfWriteWord(fp,fh.bfType); 
    czfWriteDWord(fp,fh.bfSize); 
    czfWriteWord(fp,fh.bfReserved1);
    czfWriteWord(fp,fh.bfReserved2);
    czfWriteDWord(fp,fh.bfOffBits);

    czsBmpInfoHeader ih; 
    ih.biSize = 40; 
    ih.biWidth = img.w; 
    ih.biHeight = img.h; 
    ih.biPlanes = 1; 
    ih.biBitCount = 24; 
    ih.biCompression = 0;   // BI_RGB, no compression 
    ih.biSizeImage = 0; 
    ih.biXPelsPerMeter = 0; 
    ih.biYPelsPerMeter = 0; 
    ih.biClrUsed = 0; 
    ih.biClrImportant = 0; 
    czfWriteDWord(fp,ih.biSize); 
    czfWriteDWord(fp,(unsigned int)ih.biWidth); 
    czfWriteDWord(fp,(unsigned int)ih.biHeight); 
    czfWriteWord(fp,ih.biPlanes); 
    czfWriteWord(fp,ih.biBitCount); 
    czfWriteDWord(fp,ih.biCompression); 
    czfWriteDWord(fp,ih.biSizeImage);
    czfWriteDWord(fp,(unsigned int)ih.biXPelsPerMeter); 
    czfWriteDWord(fp,(unsigned int)ih.biYPelsPerMeter);
    czfWriteDWord(fp,ih.biClrUsed);
    czfWriteDWord(fp,ih.biClrImportant); 

    // writing data 
    for (int row=img.h-1; row>=0; row--) 
    {
        int size = img.w*3; 
        unsigned char *p = &img.data[row*size]; 
        unsigned char pix[3];   // a pixel, BGR 
        unsigned char pad[4] = {0,0,0,0}; 
        for (int i=0; i<img.w; i++,p+=3) 
        {
            pix[0]=p[2], pix[1]=p[1], pix[2]=p[0];  // swap RGB to BGR 
            int writesize = fwrite((void *)pix, 1, 3, fp); 
            if (writesize != 3) 
            {
                printf ("Unable to write BMP image data!\n"); 
                fclose(fp); 
                return 0; 
            }
        }
        if (padwidth != size)   // need padding
        {
            size = padwidth - size; 
            int writesize = fwrite((void *)pad, 1, size, fp); 
            if (writesize != size) 
            {
                printf ("Unable to write BMP image data!\n"); 
                fclose(fp); 
                return 0; 
            }
        }
    }
    fclose(fp);
    return (&img); 
}

// load BMP grayscale image 
czsImgGry *czfLoadImgBMP(czsImgGry &img, const char *filename, int flag)
{
    FILE *fp; 
    if ((fp=fopen(filename, "rb")) == NULL)
    {
        printf ("Unable to open image file: %s!\n", filename); 
        return 0; 
    }
    int width, height, bpp; 
    if (czfReadBMPHeader(fp, width, height,bpp)) 
    {
        printf ("Unable to read the BMP image header for file: %s!\n", filename); 
        fclose (fp); 
        return 0; 
    }
    if (bpp != 8) 
    {
        printf ("The BMP image is not grayscale!\n"); 
        fclose (fp);    
        return 0; 
    }
    if (flag == CZ_IMAGE_MEM_NOT_ALLOCATED) 
    {
        if (!czfCreateImg(img, width, height)) 
        {
            fclose (fp); 
            return 0; 
        }
    }
    else 
    {
        if (img.w != width || img.h != height) 
        {
            printf ("The size of the image does not match with that of the image file %s!\n", filename); 
            fclose (fp); 
            return 0; 
        }
    }

    // skip the color map 
    fseek(fp, 256*4, SEEK_CUR);    // 8 bit color map, 256 of 4-byte RGBQUAD, ignored!

    // now read the data row by row, notice BMP stores pixels bottom up 
    for (int row=height-1; row>=0; row--) 
    {
        int size = width; 
        unsigned char *p = &img.data[row*size]; 
        int readsize = fread((void *)p, 1, (size_t)size, fp); 
        if (readsize != size) 
        {
            printf ("Error while reading image data! \n"); 
            if (flag == CZ_IMAGE_MEM_NOT_ALLOCATED) 
                czfReleaseImg(img); 
            fclose(fp); 
            return 0; 
        }
        // BMP has padding for each row such that each row is 4-byte aligned 
        unsigned char pad[4];   // no more than 4 bytes of padding 
        if (size%4) 
        {
            size = 4-size%4; 
            readsize = fread((void *)pad, 1, (size_t)size, fp); 
            if (readsize != size) 
            {
                printf ("Error while reading image data! \n"); 
                if (flag == CZ_IMAGE_MEM_NOT_ALLOCATED) 
                    czfReleaseImg(img); 
                fclose(fp); 
                return 0; 
            }
        }
    }    
    fclose(fp);
    return (&img); 
}

// write BMP grayscale image 
const czsImgGry *czfWriteImgBMP(const czsImgGry &img, const char *filename)
{
    FILE *fp; 
    if ((fp=fopen(filename, "wb")) == NULL)
    {
        printf ("Unable to open file %s for writing!\n", filename); 
        return 0; 
    }
    
    int padwidth = (img.w*8+31)/32*4;  // width padded to 4 bytes 
    czsBmpFileHeader fh; 
    fh.bfType = 0x4d42; 
    fh.bfSize = 14+40+256*4+padwidth*img.h; 
    fh.bfReserved1 = 0; 
    fh.bfReserved2 = 0; 
    fh.bfOffBits = 14+40+256*4; 
    czfWriteWord(fp,fh.bfType); 
    czfWriteDWord(fp,fh.bfSize); 
    czfWriteWord(fp,fh.bfReserved1);
    czfWriteWord(fp,fh.bfReserved2);
    czfWriteDWord(fp,fh.bfOffBits);

    czsBmpInfoHeader ih; 
    ih.biSize = 40; 
    ih.biWidth = img.w; 
    ih.biHeight = img.h; 
    ih.biPlanes = 1; 
    ih.biBitCount = 8; 
    ih.biCompression = 0;   // BI_RGB, no compression 
    ih.biSizeImage = 0; 
    ih.biXPelsPerMeter = 0; 
    ih.biYPelsPerMeter = 0; 
    ih.biClrUsed = 0; 
    ih.biClrImportant = 0; 
    czfWriteDWord(fp,ih.biSize); 
    czfWriteDWord(fp,(unsigned int)ih.biWidth); 
    czfWriteDWord(fp,(unsigned int)ih.biHeight); 
    czfWriteWord(fp,ih.biPlanes); 
    czfWriteWord(fp,ih.biBitCount); 
    czfWriteDWord(fp,ih.biCompression); 
    czfWriteDWord(fp,ih.biSizeImage);
    czfWriteDWord(fp,(unsigned int)ih.biXPelsPerMeter); 
    czfWriteDWord(fp,(unsigned int)ih.biYPelsPerMeter);
    czfWriteDWord(fp,ih.biClrUsed);
    czfWriteDWord(fp,ih.biClrImportant); 

    // write color map table 
    unsigned char rgbquad[4]; 
    rgbquad[3] = 0; 
    for (int i=0; i<256; i++) 
    {
        rgbquad[0] = rgbquad[1] = rgbquad[2] = i; 
        int writesize = fwrite((void *)rgbquad, 1, 4, fp); 
        if (writesize != 4) 
        {
            printf ("Unable to write BMP image data!\n"); 
            fclose(fp); 
            return 0; 
        }
    }

    // writing data 
    for (int row=img.h-1; row>=0; row--) 
    {
        int size = img.w; 
        unsigned char *p = &img.data[row*size]; 
        unsigned char pad[4] = {0,0,0,0}; 
        int writesize = fwrite((void *)p, 1, (size_t)size, fp); 
        if (writesize != size) 
        {
            printf ("Unable to write BMP image data!\n"); 
            fclose(fp); 
            return 0; 
        }
        if (padwidth != size)   // need padding
        {
            size = padwidth - size; 
            writesize = fwrite((void *)pad, 1, size, fp); 
            if (writesize != size) 
            {
                printf ("Unable to write BMP image data!\n"); 
                fclose(fp); 
                return 0; 
            }
        }
    }
    fclose(fp);
    return (&img); 
}

// retrieve color pixel value at an non-integer point 
double *czfGetPixel(const czsImgClr &img, double w, double h, double clr[3])
{
    int wl = (int)floor(w); 
    int hl = (int)floor(h); 
    if (wl!=w && hl!=h)     // happen most of the time 
    {
        if (wl<0 || hl<0 || wl>img.w-2 || hl>img.h-2) 
        {
            clr[0] = clr[1] = clr[2] = -1.0; 
            return 0; 
        }
        double alpha = w-wl; 
        double beta = 1-alpha; 
        double val1[3], val2[3]; 
        unsigned char *p = &img.data[(hl*img.w+wl)*3]; 
        val1[0] = p[0]*beta+p[3]*alpha; 
        val1[1] = p[1]*beta+p[4]*alpha; 
        val1[2] = p[2]*beta+p[5]*alpha; 
        p += img.w; 
        val2[0] = p[0]*beta+p[3]*alpha; 
        val2[1] = p[1]*beta+p[4]*alpha; 
        val2[2] = p[2]*beta+p[5]*alpha; 
        alpha = h-hl; 
        beta = 1-alpha; 
        clr[0] = val1[0]*beta+val2[0]*alpha; 
        clr[1] = val1[1]*beta+val2[1]*alpha; 
        clr[2] = val1[2]*beta+val2[2]*alpha; 
        return (clr);
    }
    else if (wl==w && hl!=h) 
    {
        if (wl<0 || hl<0 || wl>img.w-1 || hl>img.h-2)
        {
            clr[0] = clr[1] = clr[2] = -1.0; 
            return 0;             
        }
        double alpha = h-hl; 
        double beta = 1-alpha; 
        unsigned char *p1 = &img.data[(hl*img.w+wl)*3]; 
        unsigned char *p2 = p1 + img.w; 
        clr[0] = p1[0]*beta+p2[0]*alpha; 
        clr[1] = p1[1]*beta+p2[1]*alpha; 
        clr[2] = p1[2]*beta+p2[2]*alpha; 
        return (clr); 
    }
    else if (wl!=w && hl==h)
    {
        if (wl<0 || hl<0 || wl>img.w-2 || hl>img.h-1) 
        {
            clr[0] = clr[1] = clr[2] = -1.0; 
            return 0;             
        }
        double alpha = w-wl; 
        double beta = 1-alpha; 
        unsigned char *p = &img.data[(hl*img.w+wl)*3]; 
        clr[0] = p[0]*beta+p[3]*alpha; 
        clr[1] = p[1]*beta+p[4]*alpha; 
        clr[2] = p[2]*beta+p[5]*alpha; 
        return (clr); 
    }
    else
    {
        if (wl<0 || hl<0 || wl>img.w-1 || hl>img.h-1) 
        {
            clr[0] = clr[1] = clr[2] = -1.0; 
            return 0;             
        }
        unsigned char *p = &img.data[(hl*img.w+wl)*3]; 
        clr[0] = p[0], clr[1] = p[1], clr[2] = p[2]; 
        return (clr); 
    }
}

// retrieve grayscale pixel value at an integer point 
double *czfGetPixel(const czsImgGry &img, double w, double h, double &gry)
{
    int wl = (int)floor(w); 
    int hl = (int)floor(h); 
    if (wl!=w && hl!=h)     // happen most of the time 
    {
        if (wl<0 || hl<0 || wl>img.w-2 || hl>img.h-2) 
        {
            gry = -1.0; 
            return 0; 
        }
        double alpha = w-wl; 
        double val1, val2; 
        unsigned char *p = &img.data[hl*img.w+wl]; 
        val1 = p[0]*(1-alpha)+p[1]*alpha; 
        p += img.w; 
        val2 = p[0]*(1-alpha)+p[1]*alpha; 
        alpha = h-hl; 
        gry = val1*(1-alpha)+val2*alpha; 
        return (&gry);
    }
    else if (wl==w && hl!=h) 
    {
        if (wl<0 || hl<0 || wl>img.w-1 || hl>img.h-2)
        {
            gry = -1.0; 
            return 0;             
        }
        double alpha = h-hl; 
        unsigned char *p = &img.data[hl*img.w+wl]; 
        gry = (*p)*(1-alpha)+(*(p+img.w))*alpha; 
        return (&gry); 
    }
    else if (wl!=w && hl==h)
    {
        if (wl<0 || hl<0 || wl>img.w-2 || hl>img.h-1) 
        {
            gry = -1.0; 
            return 0;             
        }
        double alpha = w-wl; 
        unsigned char *p = &img.data[hl*img.w+wl]; 
        gry = p[0]*(1-alpha)+p[1]*alpha; 
        return (&gry); 
    }
    else
    {
        if (wl<0 || hl<0 || wl>img.w-1 || hl>img.h-1) 
        {
            gry = -1.0; 
            return 0;             
        }
        gry = img.data[hl*img.w+wl]; 
        return (&gry); 
    }
}

void czfInitClrConvert()
{
	using namespace cz_image_utils;  
    if (init) return;       // the table has already been initialized 
    for (int i=0; i<256; i++)
        fast_uchar[i] = 0; 
	int i = 0;
    for (i=256; i<512; i++) 
        fast_uchar[i] = i-256; 
    for (i=512; i<768; i++) 
        fast_uchar[i] = 255; 

    for (i=0; i<256; i++) 
    {
        fast_rgb2ycc[i+r_y] = ((int)(0.299*(1<<shift)+0.5))*i; 
        fast_rgb2ycc[i+g_y] = ((int)(0.587*(1<<shift)+0.5))*i; 
        fast_rgb2ycc[i+b_y] = ((int)(0.114*(1<<shift)+0.5))*i + (1<<(shift-1)); 
        fast_rgb2ycc[i+r_cb] = -((int)(0.16874*(1<<shift)+0.5))*i; 
        fast_rgb2ycc[i+g_cb] = -((int)(0.33136*(1<<shift)+0.5))*i; 
        // we use a rounding fudge-factor of 0.5-epsilon for Cb and Cr, so we can save rounding cost 
        fast_rgb2ycc[i+b_cb] = ((int)(0.5*(1<<shift)+0.5))*i + (1<<(shift-1)) - 1; 
        // fast_rgb2ycc_table[i+r_cr] = ((int)(0.5*(1<<shift)+0.5))*i + (1<<(shift-1)) - 1; // same as b_cb
        fast_rgb2ycc[i+g_cr] = -((int)(0.41869*(1<<shift)+0.5))*i; 
        fast_rgb2ycc[i+b_cr] = -((int)(0.08131*(1<<shift)+0.5))*i; 

        fast_ycc2rgb[i+cr_r] = (((int)(1.402*(1<<shift)+0.5))*(i-128) + (1<<(shift-1)))>>shift; 
        fast_ycc2rgb[i+cb_g] = ((int)(0.34414*(1<<shift)+0.5))*(i-128) + (1<<(shift-1)); 
        fast_ycc2rgb[i+cr_g] = ((int)(0.71414*(1<<shift)+0.5))*(i-128); 
        fast_ycc2rgb[i+cb_b] = (((int)(1.772*(1<<shift)+0.5))*(i-128) + (1<<(shift-1)))>>shift; 
    }
    init = 1;   // flag to show that the tables have been initialized 
}

// convert a color image to grayscale image 
czsImgGry *czfClrImgToGry(const czsImgClr &imgc, czsImgGry &imgg)
{   
    if (imgc.w != imgg.w || imgc.h != imgg.h) 
    {
        printf ("Unmatched image size when performing color conversion!\n"); 
        return 0; 
    }

    czfInitClrConvert(); 
    int size = imgg.w*imgg.h; 
    unsigned char *pc = imgc.data; 
    unsigned char *pg = imgg.data; 
    for (int i=0; i<size; i++) 
    {
        CZ_RGB2GRY(pc, pg); 
        pg ++; 
        pc += 3; 
    }
    return (&imgg); 
}


// convert a grayscale image to a color image 
czsImgClr *czfGryImgToClr(const czsImgGry &imgg, czsImgClr &imgc)
{
    if (imgc.w != imgg.w || imgc.h != imgg.h) 
    {
        printf ("Unmatched image size when performing color conversion!\n"); 
        return 0; 
    }

    czfInitClrConvert(); 
    int size = imgg.w*imgg.h; 
    unsigned char *pc = imgc.data; 
    unsigned char *pg = imgg.data; 
    for (int i=0; i<size; i++) 
    {
        CZ_GRY2RGB(pg, pc); 
        pg ++; 
        pc += 3; 
    }
    return (&imgc); 
}
