/********************************************************************************
 * Copyright (c) 2004, Cha Zhang <czhang@ieee.org>, Advanced Multimedia 
 * Processing Lab, ECE, Carnegie Mellon University. 
 * All rights reserved. 
 * 
 * History
 * -------
 *
 * 2004.5.15 Created by Cha Zhang at CMU as a simplified version of the
 *           rendering algorithm implemented for the mobile camera array
 *           at AMP Lab, ECE, Carnegie Mellon University. 
 ********************************************************************************/

/********************************************************************************
 * This is the main routine of CAView, which stands for 'camera array viewer'. 
 * The algorithm is detailed in the paper: 
 *     Cha Zhang and Tsuhan Chen, "A Self-Reconfigurable Camera Array", 2004 
 *     Eurographics Symposium on Rendering, June, 2004. 
 ********************************************************************************/

#ifdef _WIN32
#include "windows.h" 
#endif

#include "stdio.h"
#include "stdlib.h" 
#include "string.h"
#include "time.h"
#include "GL/gl.h"
#include "GL/glu.h"
#include <GL/glut.h>
#include "cz_math_utils.h" 
#include "cz_image_utils.h"
#include "cz_calib_utils.h"
#include "cz_quatcam_utils.h"


//GLUI stuff
#include <GLUI/src/glui.h>
int window;
int globalButton;
float matrixR [16] = {1,0,0,0, 0,1,0,0, 0,0,1,0, 0,0,0,1}; // matrix used for GLUI trackball rotation
float matrixT [3] = {0,0,0}; // matrix used for GLUI translation
float old_matrixT[3] = {0,0,0};
int gui_depth; //used to change the depth in constant depth mode
int depth_change = 0;
int old_depth_increase = 0;
int gui_depth_mode = 0;
int gui_view = 0;


#define MAX_NUM_INTERP  4       // maximum number of light rays used for interpolation 

// Below are parameters loaded from the *.cav scene file 
static int numimg;              // number of scene images
static int width;               // width of the input images 
static int height;              // height of the input images
static double fov;              // horizontal field of view 
static double mindepth;         // minimum depth of the scene 
static double maxdepth;         // maximum depth of the scene
static double dftdepth;         // default depth of the scene for constant rendering 
static int numdepth;            // number of depth layer for hypothesis testing 
static int depthlevel;          // maximum depth subdivision level 
static int meshsize;            // size of the initial 2D mesh, in terms of pixels 
static int meshlevel;           // maximum subdivision level from the initial 2D mesh 
static int wndsize;             // patch window size for verifying the color consistency 
static int wndstep;             // subsampling stepsize of the patch window 
static int ccv_mode;            // mode to perform color consistency verification 
static int avgdiffth;           // threshold used for early rejection during color consistency verification using COR 
static int sum2th;              // threshold used for early acceptance during color consistency verification using COR 
static int numinterp;           // number of light rays for interpolation 

// the images and their calibration parameters 
static czsImgClr *img = NULL;   // image array
static czsCalibPara *cp = NULL; // calibration parameter array
static int *imgused = NULL;     // array indicating whether an image is used during the rendering of the current view 

// about texture 
static GLuint *texname = NULL;  // texture name (id) array 
static int texwidth; 
static int texheight; 

// representation of the vertices of the 2D mesh for both depth recovery and rendering 
static int numvtx_wd;           // number of vertices at the densest subdivision level along width
static int numvtx_hd;           // number of vertices at the densest subdivision level along height 
static int numvtx_wc;           // number of vertices at the coarsest subdivision level along width
static int numvtx_hc;           // number of vertices at the coarsest subdivision level along height
struct Vertex            
{
    czsPt2 pos;         // the postion of the vertices on the virtual view's imaging plane (z=-1)
    czsVec3 dir;        // light ray direction in the world coordinate 
    int imgid[MAX_NUM_INTERP];          // indices of the images used for rendering this vertex
    double weight[MAX_NUM_INTERP];      // weight of the images used for interpolating this sample 
    czsPt2 projpos[MAX_NUM_INTERP];     // projection position to the close-by cameras (2D pixel position)
    int depidx;         // depth index of the sample 
    double val;         // the consistency score value 
}; 
static Vertex *vtx = NULL;              // vertex array 

// the triangles used for processing and rendering 
struct Triangle
{
    int v[3]; 
    int level;          // subdivision level
}; 
static Triangle *tri1 = NULL;        
static Triangle *tri2 = NULL;        // processing triangle arrays
static int numtri1; 
static int numtri2; 

// things related with the virtual rendering camera 
static czsQuatCam rqcam;                // quaternion camera of rendering 
static czsPt3 rndpos;                   // the rendering position in world coordinate space 
static czsVec3 allcam_ctr;              // center of all the capturing cameras in the world coordinate space 
static double stepfwd;                  // step size of depth change 
static double stepside;                 // step size of translations/sidesteps

// parameters used for depth recovery 
static double constdepth;               // constant depth for rendering 
static double *depth;                   // test depth array distributed evenly in disparity space 
static double *sdepth;                  // subdivided test depth array 
static czsVec3 xdir, ydir, zdir;        // vectors used for transforming virtual view coordinates to world coordinates
static int *imglist = NULL;             // intermediate array to store image id's 
static double *dist = NULL;             // intermediate array to store distances 
static int *flag = NULL;                // intermediate array to store flags 
static double *pixval = NULL;           // intermediate array to store pixel values 
static double *val = NULL;              // intermediate array to store consistency score values 
static czsPt2 **projpos = NULL;         // intermediate array to store vertex projection positions 
static int min_depidx, max_depidx;      // minimum and maximum depth index for a triangle

// constant depth or use our depth estimation algorithm
#define CONST_DEPTH         0
#define RECON_DEPTH         1
static int depth_mode = RECON_DEPTH;    // rendering depth mode, default as reconstructed depth 

// different rendering mode 
#define RND_SCENE           1
#define RND_FRAME           2
#define RND_BOTH            3
static int rnd_mode = RND_SCENE;        // rendering mode, default only show the scene 

// different color consistency verification mode 
#define CCV_SSD             1
#define CCV_COR             2

// time variables to get the time used for rendering
double startTime = 0;
double endTime = 0;





// provides a clean exit of the program 
void exit_caview(char *msg = NULL)
{
    if (msg != NULL) 
        printf ("%s\n", msg); 
    if (img != NULL) 
    {
        for (int i=0; i<numimg; i++) 
            czfReleaseImg(img[i]); 
        free(img); 
    }
    if (cp != NULL) free(cp); 
    if (texname != NULL) free(texname); 
    if (vtx != NULL) free(vtx); 
    if (tri1 != NULL) free(tri1); 
    if (tri2 != NULL) free(tri2); 
    if (depth != NULL) free(depth); 
    if (sdepth != NULL) free(sdepth); 
    if (imglist != NULL) free(imglist); 
    if (dist != NULL) free(dist);
    if (flag != NULL) free(flag); 
    if (pixval != NULL) free(pixval); 
    if (val != NULL) free(val); 
    if (projpos != NULL) 
    {
        for (int i=0; i<numinterp; i++) 
        {
            if (projpos[i] != NULL) 
                free(projpos[i]); 
        }
        free(projpos); 
    }
    exit (0); 
}


// load the scene files based on a given configuration file 
int loadscene (char *fname) 
{
    // make sure we are loading a *.cav file 
    char *suffix = strrchr(fname, '.'); 
    if (strcmp(suffix, ".cav") != 0) 
    {
        printf ("Only '*.cav' files can be loaded with this program!\n"); 
        return 0; 
    }
    // remove the file name and get the path 
    char *str1 = strrchr (fname, '/'); 
    char *str2 = strrchr (fname, '\\');     // this might be the case for windows systems 
    char pathname[256]=""; 
    if (str1 == NULL && str2 == NULL) 
        ;       // pathname remains empty 
    else if (str1 == NULL) 
        strncpy(pathname, fname, strlen(fname)-strlen(str2)+1); 
    else if (str2 == NULL) 
        strncpy(pathname, fname, strlen(fname)-strlen(str1)+1); 
    else 
        strncpy(pathname, fname, strlen(fname)-(CZ_MIN2(strlen(str1),strlen(str2)))+1); 

    FILE *fp; 
    if ((fp = fopen(fname, "r")) == NULL) 
    {
        printf ("Unable to open scene file %s\n", fname);
        return 0; 
    }
    char imgname[128];      // prefix of the image file names 
    char calibname[128];    // prefix of the camera calibration parameter file names 
    fscanf (fp, "NUM_IMG=%d\n", &numimg); 
    fscanf (fp, "IMG_FNAME=%s\n", imgname); 
    fscanf (fp, "CALIB_FNAME=%s\n", calibname); 
    fscanf (fp, "WIDTH=%d\n", &width); 
    fscanf (fp, "HEIGHT=%d\n", &height); 
    fscanf (fp, "FOV=%lf\n", &fov); 
    fscanf (fp, "MIN_DEPTH=%lf\n", &mindepth); 
    fscanf (fp, "MAX_DEPTH=%lf\n", &maxdepth);   
    fscanf (fp, "DFT_DEPTH=%lf\n", &dftdepth); 
    fscanf (fp, "NUM_DEPTH=%d\n", &numdepth); 
    fscanf (fp, "DEPTH_SUBDIV_LEVEL=%d\n", &depthlevel); 
    fscanf (fp, "INIT_MESH_SIZE=%d\n", &meshsize); 
    fscanf (fp, "MESH_SUBDIV_LEVEL=%d\n", &meshlevel); 
    fscanf (fp, "WND_SIZE=%d\n", &wndsize); 
    fscanf (fp, "WND_STEP=%d\n", &wndstep); 
    fscanf (fp, "CCV_MODE=%d\n", &ccv_mode); 
    fscanf (fp, "AVG_DIFF_TH=%d\n", &avgdiffth); 
    fscanf (fp, "SUM2_TH=%d\n", &sum2th); 
    fscanf (fp, "NUM_INTERP=%d\n", &numinterp); 
    fclose (fp); 

    if (numinterp > MAX_NUM_INTERP)
    {
        printf ("Warning: number of images used for interpolation set to maximum value %d\n", MAX_NUM_INTERP); 
        numinterp = MAX_NUM_INTERP; 
    }
    if (ccv_mode != CCV_SSD && ccv_mode != CCV_COR) 
    {
        printf ("Warning: unrecognized color consistency verification mode! Set to SSD.\n"); 
        ccv_mode = CCV_SSD;
     }

    fov *= CZ_D2R;          // convert degree to radian
    constdepth = dftdepth;  // if we perform constant depth rendering, use dftdepth initially 

    // allocate memory for image structures  
    img = (czsImgClr *)calloc(numimg, sizeof(czsImgClr)); 
    if (img == NULL) 
    {
        printf ("Unalbe to allocate memory for the images!\n"); 
        return 0; 
    }
    cp = (czsCalibPara *)malloc(numimg*sizeof(czsCalibPara)); 
    if (cp == NULL) 
    {
        printf ("Unable to allocate memory for calibration parameters!\n"); 
        return 0; 
    }
    printf ("Loading images and calibration parameters..."); 
    czfV3SetZero(allcam_ctr); 
    for (int i=0; i<numimg; i++) 
    {
        // load the image 
        char fullname[512]; 
        // here we assume the input images are in PPM format
        sprintf(fullname, "%s%s%04d.ppm", pathname, imgname, i); 
        if (czfLoadImgPPM(img[i], fullname) == 0) 
            return 0; 
        // notice image.cpp also support BMP files, you may use the following 3 lines instead
//        sprintf(fullname, "%s%s%04d.bmp", pathname, imgname, i); 
//        if (czfLoadImgBMP(img[i], fullname) == 0) 
//            return 0; 
        if (img[i].w != width || img[i].h != height) 
        {
            printf ("Image %s has different size as specified in %s\n", fullname, fname); 
            return 0; 
        }

        // load camera calibration parameters 
        sprintf(fullname, "%s%s%04d.txt", pathname, calibname, i); 
        if ((fp = fopen(fullname, "r")) == NULL) 
        {
            printf ("Unable to load calibration file %s.\n", fullname); 
            return 0; 
        }
        fscanf (fp, "fc = [ %lf ; %lf ];\n", &cp[i].fc[0], &cp[i].fc[1]); 
        fscanf (fp, "cc = [ %lf ; %lf ];\n", &cp[i].cc[0], &cp[i].cc[1]); 
        fscanf (fp, "alpha_c = %lf;\n", &cp[i].alpha_c); 
        fscanf (fp, "kc = [ %lf ; %lf ; %lf ; %lf ; %lf ];\n", 
                &cp[i].kc[0], &cp[i].kc[1], &cp[i].kc[2], &cp[i].kc[3], &cp[i].kc[4]);
        fscanf (fp, "Tc_ext = [ %lf \t %lf \t %lf ];\n", &cp[i].T.x, &cp[i].T.y, &cp[i].T.z); 
        fscanf (fp, "Rc_ext = [ %lf \t %lf \t %lf\n", &cp[i].R.m[0], &cp[i].R.m[1], &cp[i].R.m[2]); 
        fscanf (fp, "           %lf \t %lf \t %lf\n", &cp[i].R.m[3], &cp[i].R.m[4], &cp[i].R.m[5]); 
        fscanf (fp, "           %lf \t %lf \t %lf ];\n", &cp[i].R.m[6], &cp[i].R.m[7], &cp[i].R.m[8]); 
        fclose (fp); 

        // obtain the center of projection for each camera, also accumulate allcam_ctr
        czsMat3 invrotmat; 
        czfM3Inv(cp[i].R, invrotmat); 
        cp[i].ctr.x = -cp[i].T.x; 
        cp[i].ctr.y = -cp[i].T.y; 
        cp[i].ctr.z = -cp[i].T.z; 
        czfV3MulPtMat(cp[i].ctr, invrotmat); 
        czfV3Add(allcam_ctr, cp[i].ctr); 


        // is there any lens distortion? 
        if (cp[i].kc[0] == 0.0 && cp[i].kc[1] == 0.0 && cp[i].kc[2] == 0.0 && cp[i].kc[3] == 0.0 && cp[i].kc[4] == 0.0) 
        {   
			// no lens distortion 
            cp[i].no_lens_distortion = 1; 
            // construct the 3x4 projection matrix (combining internal and externl matrices) 
            double *p = cp[i].P.m; 
            double *r = cp[i].R.m; 
            double *fc = cp[i].fc; 
            double *cc = cp[i].cc; 
            double alpha_c = cp[i].alpha_c; 
            p[0] = fc[0]*r[0]+alpha_c*fc[0]*r[3]+cc[0]*r[6]; 
            p[1] = fc[0]*r[1]+alpha_c*fc[0]*r[4]+cc[0]*r[7]; 
            p[2] = fc[0]*r[2]+alpha_c*fc[0]*r[5]+cc[0]*r[8]; 
            p[3] = fc[0]*cp[i].T.x+cp[i].alpha_c*fc[0]*cp[i].T.y+cc[0]*cp[i].T.z; 
            p[4] = fc[1]*r[3]+cc[1]*r[6]; 
            p[5] = fc[1]*r[4]+cc[1]*r[7]; 
            p[6] = fc[1]*r[5]+cc[1]*r[8]; 
            p[7] = fc[1]*cp[i].T.y+cc[1]*cp[i].T.z; 
            p[8] = r[6]; 
            p[9] = r[7]; 
            p[10] = r[8]; 
            p[11] = cp[i].T.z; 
        }
        else
            cp[i].no_lens_distortion = 0; 

        if (i%2 == 0) 
        {
            printf (".");     // show some progress in stdout
            fflush(stdout); 
        }
    }
    printf ("\n");
    czfV3MulScalar(allcam_ctr, 1.0/numimg);         // this is the center of all the capturing cameras 
    stepfwd = (dftdepth-allcam_ctr.z) * 0.01;
    stepside = (mindepth-allcam_ctr.z)/cp[0].fc[0]; 

    return 1; 
}

// initialize the rendering process 
void init_rendering () 
{
    // in OpenGL, the texture must have width and height as 2 to the power of something
    texwidth = czfRoundUpPow2(width); 
    texheight = czfRoundUpPow2(height); 

    // create the textures with a dummy image of size texwidth*texheight
    printf ("Initializing texture buffers...\n");     
    texname = (GLuint *)malloc(numimg*sizeof(GLuint)); 
    if (texname == NULL) 
        exit_caview("Unable to allocate memory for texture names!\n"); 
    glGenTextures (numimg, texname); 
    GLubyte *dummybuffer = (GLubyte *)calloc(texwidth*texheight*3,sizeof(GLubyte)); 
    if (dummybuffer == NULL) 
        exit_caview("Unable to allocate memory for creating the texture maps!\n"); 
    glPixelStorei (GL_UNPACK_ALIGNMENT, 1);
    for (int i=0; i<numimg; i++)
    {
        glBindTexture(GL_TEXTURE_2D, texname[i]); 
        glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, texwidth, texheight, 0, GL_RGB, GL_UNSIGNED_BYTE, dummybuffer); 
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP); 
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP); 
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR); 
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR); 
        // the input image is put in the center part of the texture map 
        glTexSubImage2D(GL_TEXTURE_2D, 0, (texwidth-width)/2, (texheight-height)/2, width, height, GL_RGB, GL_UNSIGNED_BYTE, img[i].data); 
    }
    free(dummybuffer); 

    printf ("Allocating memory for vertices and triangles...\n"); 
    // allocate memory for vertices 
    numvtx_wc = width/meshsize+1; 
    numvtx_hc = height/meshsize+1; 
    numvtx_wd = (width/meshsize)*(1<<meshlevel)+1; 
    numvtx_hd = (height/meshsize)*(1<<meshlevel)+1; 
    vtx = (Vertex *)malloc((numvtx_wd*numvtx_hd)*sizeof(Vertex)); 
    if (vtx == NULL) 
        exit_caview("Unable to allocate memory for the vertices!\n"); 

    // obtain the postion of the vertices on the virtual view's imaging plane 
    double w_start = -tan(fov/2); 
    double h_start = w_start*height/width; 
    double w_step = -2*w_start/(numvtx_wd-1); 
    double h_step = -2*h_start/(numvtx_hd-1); 
    int k = 0; 
	int i = 0;							//piin
    for (i=0; i<numvtx_hd; i++) 
        for (int j=0; j<numvtx_wd; j++) 
        {
            vtx[k].pos.x = w_start + j*w_step; 
            vtx[k++].pos.y = h_start + i*h_step; 
        }
    
    // make sure we have enough memory allocated for triangles 
    // at the most detailed subdivision level, number of *rectangles* is less than numvtx_wd*numvtx_hd, 
    // so we have no more than numvtx_wd*numvtx_hd*2 triangles at this level
    // consider there are multiple subdivision levels, we scale the above number by 4/3, which is 1+1/4+1/16+...
    tri1 = (Triangle *)malloc((numvtx_wd*numvtx_hd)*2*4*sizeof(Triangle)/3);
    tri2 = (Triangle *)malloc((numvtx_wd*numvtx_hd)*2*4*sizeof(Triangle)/3); 
    numtri1 = 0; 
    numtri2 = 0; 
    if (tri1 == NULL || tri2 == NULL) 
        exit_caview("Unable to allocate memory for the triangles!\n"); 

    // allocate memory for the test plane depths 
    depth = (double *)malloc(numdepth*sizeof(double)); 
    sdepth = (double *)malloc(((numdepth-1)*(1<<depthlevel)+1)*sizeof(double)); 
    if (depth == NULL || sdepth == NULL) 
        exit_caview("Unable to allocate memory for depth arrays!\n"); 

    // allocate memory for intermediate arrays 
    imgused = (int *)malloc(numimg*sizeof(int)); 
    imglist = (int *)malloc(numimg*sizeof(int)); 
    dist = (double *)malloc(numimg*sizeof(double)); 
    flag = (int *)malloc(numimg*sizeof(int)); 
    pixval = (double *)malloc(numinterp*wndsize*wndsize*sizeof(double)); 
    val = (double *)malloc((numdepth+2)*sizeof(double)); 
    projpos = (czsPt2 **)calloc(numinterp,sizeof(czsPt2 *)); 
    if (imgused == NULL || imglist == NULL || dist == NULL || flag == NULL || 
        pixval == NULL || val == NULL || projpos == NULL) 
        exit_caview("Unable to allocate memory for intermediate arrays!\n"); 
    for (i=0; i<numinterp; i++) 
    {
        projpos[i] = (czsPt2 *)malloc(((numdepth-1)*(1<<depthlevel)+1)*sizeof(czsPt2)); 
        if (projpos[i] == NULL) 
            exit_caview("Unable to allocate memory for projection position array!\n"); 
    }

    // initialize the rendering camera 
    czsPt3 initpos; 
    czsVec3 initaxis; 
    double aspratio = double(width)/height; 
    czfV3SetZero(initpos); 
    czfV3Set (initaxis, 0.0, 0.0, -1.0);    
    // rqcam follows OpenGl coordinate, initially the virtual camera look at -z direction 
    czfQuatCamSetInit(rqcam, initpos, 0.0, initaxis, 2*atan(tan(fov/2)/aspratio)); 
    czfQuatCamReset(rqcam); 

    // get the virtual camera position in the world coordinate 
    // initally, we put the virtual camera at the center of all the capturing cameras 
    // notice rqcam is in OpenGL coordinate, rndpos is in our world coordinate 
    rndpos.x = rqcam.pos.x + allcam_ctr.x; 
    rndpos.y = -rqcam.pos.y + allcam_ctr.y; 
    rndpos.z = -rqcam.pos.z + allcam_ctr.z; 

    // initialize OpenGL stuff 
    glViewport (0, 0, width, height); 
    glMatrixMode(GL_PROJECTION); 
    glLoadIdentity();
    gluPerspective(2*atan(tan(fov/2)/aspratio)*CZ_R2D, aspratio, 0.1f, 100.0f);
    glClearColor (0.0, 0.0, 0.0, 1.0); 
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity(); 
}

// keyboard operations 
void keyboardfunc (unsigned char key, int x, int y)
{
    switch (key) 
    {
    case 27:    // 'Esc' key, will clean up the memory and exit the program
        exit_caview(); 
        break; 
    case 'r':
    case 'R': 
        czfQuatCamReset(rqcam); 
        constdepth = dftdepth; 
        // locate the rendering position
        // notice rqcam is in OpenGL coordinate, rndpos is in our world coordinate 
        rndpos.x = rqcam.pos.x + allcam_ctr.x; 
        rndpos.y = -rqcam.pos.y + allcam_ctr.y; 
        rndpos.z = -rqcam.pos.z + allcam_ctr.z; 
        break; 
    case 'i':            
        if (depth_mode == CONST_DEPTH)
        {
            constdepth += stepfwd;
            printf ("Constant rendering depth increased to %lf\n", constdepth); 
        }
        break; 
	case 'w':
		glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
		break;
	case 'p':
		glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
		break;
    case 'I':
        if (depth_mode == CONST_DEPTH)
        {
            constdepth += 5*stepfwd;
            printf ("Constant rendering depth increased to %lf\n", constdepth); 
        }
        break; 
    case 'd':
        if (depth_mode == CONST_DEPTH) 
        {
            constdepth -= stepfwd; 
            printf ("Constant rendering depth decreased to %lf\n", constdepth); 
        }
        break;
    case 'D': 
        if (depth_mode == CONST_DEPTH) 
        {
            constdepth -= 5*stepfwd; 
            printf ("Constant rendering depth decreased to %lf\n", constdepth); 
        }
        break;
    case 't': 
    case 'T': 
        if (depth_mode == CONST_DEPTH) 
        {
            depth_mode = RECON_DEPTH; 
            printf ("Switched to rendering with reconstructed depth\n"); 
        }
        else
        {
            depth_mode = CONST_DEPTH; 
            rnd_mode = RND_SCENE; 
            printf ("Switched to rendering with constant depth\n"); 
        }
        break; 
    case 'f': 
    case 'F': 
        if (depth_mode == RECON_DEPTH) 
        {
            if (rnd_mode == RND_SCENE) 
                rnd_mode = RND_FRAME; 
            else if (rnd_mode == RND_FRAME)
                rnd_mode = RND_BOTH; 
            else
                rnd_mode = RND_SCENE; 
        }
        break; 
    case 'h': 
    case 'H': 
        printf ("\nKeyboard operation help: \nMotion:\n\t"); 
        printf ("Arrows:          translation.\n\t"); 
        printf ("Shift + Arrows:  translation with a larger step size.\n\t"); 
        printf ("Alt + Arrows:    rotation.\n\t"); 
        printf ("Home/End:        move forward/backward.\n\t"); 
        printf ("PgUp/PgDn:       move forward/backward.\n"); 
        printf ("Function:\n\t"); 
        printf ("Esc:             exit (Preferred).\n\t"); 
        printf ("h or H:          help info.\n\t"); 
        printf ("r or R:          reset to original view position and depth.\n\t"); 
        printf ("i or I:          increase the rendering depth (for constant depth rendering).\n\t"); 
        printf ("d or D:          decrease the rendering depth (for constant depth rendering).\n\t"); 
        printf ("t or T:          toggle between constant depth rendering or reconstructed depth rendering.\n\t"); 
        printf ("f or F:          toggle among showing rendered image, reconstructed 2D mesh or both.\n\t"); 
        break; 
    default: 
        break; 
    }

    glutPostRedisplay();
}

// special key operations 
void specialfunc (int key, int x, int y)
{
    czsQuat q; 
    czsQuatCam qcam; 

    switch (key)
    {
    case GLUT_KEY_LEFT:
        if (glutGetModifiers() == GLUT_ACTIVE_ALT) // rotate left 
        {
            czfV3Set (qcam.rot_axis, 0, 1, 0); 
            qcam.rot_angle = 0.05; 
            czfQuatCamGet(qcam, q); 
            czfQuatCamRotate(rqcam, q, 0); 
        }
        else if (glutGetModifiers() == GLUT_ACTIVE_SHIFT)   // move left with a large step size
            czfQuatCamTranslate(rqcam, -5*stepside, 0, 0); 
        else    // move left 
            czfQuatCamTranslate(rqcam, -stepside, 0, 0); 
        break;  
    case GLUT_KEY_RIGHT:
        if (glutGetModifiers() == GLUT_ACTIVE_ALT) // rotate right
        {
            czfV3Set (qcam.rot_axis, 0, 1, 0); 
            qcam.rot_angle = -0.05; 
            czfQuatCamGet(qcam, q); 
            czfQuatCamRotate(rqcam, q, 0); 
        }
        else if (glutGetModifiers() == GLUT_ACTIVE_SHIFT)   // move right with a large step size
            czfQuatCamTranslate(rqcam, 5*stepside, 0, 0); 
        else    // move right
            czfQuatCamTranslate(rqcam, stepside, 0, 0); 
        break;
    case GLUT_KEY_UP:
        if (glutGetModifiers() == GLUT_ACTIVE_ALT) // rotate up
        {
            czfV3Set (qcam.rot_axis, 1, 0, 0); 
            qcam.rot_angle = 0.05; 
            czfQuatCamGet(qcam, q); 
            czfQuatCamRotate(rqcam, q, 0); 
        }
        else if (glutGetModifiers() == GLUT_ACTIVE_SHIFT)   // move up with a large step size
            czfQuatCamTranslate(rqcam, 0, 5*stepside, 0); 
        else    // move up
            czfQuatCamTranslate(rqcam, 0, stepside, 0); 
        break; 
    case GLUT_KEY_DOWN:
        if (glutGetModifiers() == GLUT_ACTIVE_ALT) // rotate down
        {
            czfV3Set (qcam.rot_axis, 1, 0, 0); 
            qcam.rot_angle = -0.05; 
            czfQuatCamGet(qcam, q); 
            czfQuatCamRotate(rqcam, q, 0); 
        }
        else if (glutGetModifiers() == GLUT_ACTIVE_SHIFT)   // move down with a large step size
            czfQuatCamTranslate(rqcam, 0, -5*stepside, 0); 
        else    // move down
            czfQuatCamTranslate(rqcam, 0, -stepside, 0); 
        break; 
    case GLUT_KEY_HOME:         // move forward
    case GLUT_KEY_PAGE_UP:
        czfQuatCamTranslate(rqcam, 0, 0, -stepfwd); 
        break; 
    case GLUT_KEY_END:          // move backward
    case GLUT_KEY_PAGE_DOWN:    // increase rendering depth
        czfQuatCamTranslate(rqcam, 0, 0, stepfwd); 
        break; 
    }

    // locate the rendering position
    // notice rqcam is in OpenGL coordinate, rndpos is in our world coordinate 
    rndpos.x = rqcam.pos.x + allcam_ctr.x; 
    rndpos.y = -rqcam.pos.y + allcam_ctr.y; 
    rndpos.z = -rqcam.pos.z + allcam_ctr.z; 

    glutPostRedisplay(); 
}


// mouse function variables
int x_distance_T = 0, y_distance_T = 0;//x + y distance saved for mouse drag xy-translation
int z_distance_T = 0;//z distance saved for mouse drag z-translation
int x_distance_R = 0, y_distance_R = 0;// x + y distance saved for mouse drag xy-rotation
int new_x_R = 0, new_y_R = 0;//variable used to calculate mouse drag xy-rotation
int new_x_T = 0, new_y_T = 0;//variable used to calculate mouse drag xy-translation
int temp_old_x_T = 0, temp_old_y_T = 0;//variable used to calculate mouse drag xy-translation
int temp_old_x_R = 0, temp_old_y_R=0; //variable used to calculate mouse drag xy-rotation
int temp_old_z_T = 0; //variable used to calculate mouse drag z-translation
float rot_x = 0.0;// x-distance for mouse click rotation
float rot_y = 0.0;// y-distance for mouse click rotation
float trans_x = 0.0;// x-distance for x-axis translation
float trans_y = 0.0;// y-distnace for y-axis translation
float trans_z = 0.0;// z-distnace for z-axis translation
float old_x;
float old_y;
float old_z;
float old_x_r;
float old_y_r;
void mouseFunc(int button, int state, int x, int y)
{
//	std::cout << std::endl << "mouseFunc";
	globalButton = button;
	if (state == GLUT_DOWN && globalButton == GLUT_MIDDLE_BUTTON)
		old_z= y; // new start x point
/*	if (state == GLUT_UP && globalButton == GLUT_MIDDLE_BUTTON)
		z_distance_T = trans_z; //update total x distance
*/	
	if (state == GLUT_DOWN && globalButton == GLUT_LEFT_BUTTON)
	{
		old_x = x;
		old_y = y;
	}
	/*
	if (state == GLUT_UP && globalButton == GLUT_LEFT_BUTTON)
	{
		x_distance_T = trans_x; //update total x distance
		y_distance_T = trans_y; //update total y distance
	}	
*/
	
    if (state == GLUT_DOWN && globalButton == GLUT_RIGHT_BUTTON)
	{
		old_x_r = x;
		old_y_r = y;
	}
/*
	if (state == GLUT_UP && globalButton == GLUT_RIGHT_BUTTON)
	{
		x_distance_R = rot_x; //update total x distance
		y_distance_R = rot_y; //update total y distance
	}
*/
	 glutPostRedisplay();
}
///handle the mouse drags 
int z = 0;
void motionFunc(int x, int y)
{
	czsQuat q; 
	czsQuatCam qcam; 
	if (globalButton == GLUT_LEFT_BUTTON)
	{
		trans_x = x-old_x;
		trans_y = y-old_y;
		czfQuatCamTranslate(rqcam, stepside*trans_x, 0, 0); 
		czfQuatCamTranslate(rqcam, 0,-stepside*trans_y, 0); 
		
	}

	if (globalButton == GLUT_RIGHT_BUTTON)
	{	
		rot_x = x-old_x_r;
		rot_y = y-old_y_r;
		czfV3Set (qcam.rot_axis, 0, 1, 0); 
        qcam.rot_angle = -rot_x*0.05*0.1;
        czfQuatCamGet(qcam, q); 
        czfQuatCamRotate(rqcam, q, 0); 

		czfV3Set (qcam.rot_axis, 1, 0, 0); 
        qcam.rot_angle = -rot_y*0.05*0.1; 
        czfQuatCamGet(qcam, q); 
        czfQuatCamRotate(rqcam, q, 0); 
		
	}
	
	if (globalButton == GLUT_MIDDLE_BUTTON)
	{
		z = y;
		trans_z = z-old_z;
		czfQuatCamTranslate(rqcam, 0, 0, -stepside*trans_z);
	}

	
	rndpos.x = rqcam.pos.x + allcam_ctr.x; 
    rndpos.y = -rqcam.pos.y + allcam_ctr.y; 
    rndpos.z = -rqcam.pos.z + allcam_ctr.z; 

	old_x = x;
	old_y = y;
	old_z = z;
	old_x_r = x;
	old_y_r = y;

	glutPostRedisplay();

}


// given a light ray corresponding to a vertex *pV, find in a list of images its closeby ones 
// 'lst' is a list of image indices, 'num' is the size of the list 
void find_nearby_cameras(Vertex *pV, int *lst, int num)
{
    // get the direction of the light ray in the world coordinate space
    pV->dir.x = pV->pos.x*xdir.x + pV->pos.y*ydir.x + zdir.x; 
    pV->dir.y = pV->pos.x*xdir.y + pV->pos.y*ydir.y + zdir.y; 
    pV->dir.z = pV->pos.x*xdir.z + pV->pos.y*ydir.z + zdir.z; 
    czfV3Norm(pV->dir); 

    czsVec3 camdir, tmp; 
    for (int k=0; k<num; k++) 
    {
        flag[k] = 0;
        czfV3Sub(rndpos, cp[lst[k]].ctr, camdir); 
        czfV3Cross(camdir, pV->dir, tmp); 
        dist[k] = czfV3Len(tmp);
        // this is the distance from the capturing camera's center of projection to the light ray being considered
        // we are making use of this simplified method because we assume the cameras are roughly on a plane 
    }

    // find the nearest numinterp cameras
    int m, n=0; 
    while (n<numinterp) 
    {
        m = 0; 
        while (flag[m]) m++;    // skip those whose flags have been marked as 1
        double min_dist = dist[m]; 
        int min_m = m++; 
        for (; m<num; m++) 
        {
            if ((dist[m]<min_dist) && (!flag[m]))
            {
                min_dist = dist[m]; 
                min_m = m; 
            }
        }
        pV->imgid[n] = lst[min_m]; 
        pV->weight[n] = 1/(min_dist+CZ_SMALL_NUMBER);
        // CZ_SMALL_NUMBER is here in case min_dist is too small
        flag[min_m] = 1; 
        n++; 
    }
}

// construct the initial mesh on the virtual imaging plane 
void construct_initmesh()
{
    // obtain the virtual camera directions in the world coordinate 
    czsVec3 viewdir, updir, rightdir; 
    czfQuatCamGetOrthoDirs(rqcam, viewdir, updir, rightdir);      
    // viewdir, updir, rightdir are guaranteed to be unit vectors
    // rqcam is using OpenGL coordinate system, convert to our world coordinate system 
    xdir.x = rightdir.x, xdir.y = -rightdir.y, xdir.z = -rightdir.z; 
    ydir.x = -updir.x, ydir.y = updir.y, ydir.z = updir.z;      // ydir is along opposite direction of updir 
    zdir.x = viewdir.x, zdir.y = -viewdir.y, zdir.z = -viewdir.z; 
    
    numtri1 = 0; 
    numtri2 = 0; 
    int lev = 1<<meshlevel; 
    // initialize the triangles on the initial coarse mesh model
    for (int i=0; i<numvtx_hc-1; i++)     
        for (int j=0; j<numvtx_wc-1; j++) 
        {
            //        v1 o o o o o v2           v1 --------- v2
            //           o o o o o                 |     / |
            //           o o o o o         or      |   /   |
            //           o o o o o                 | /     |
            //        v3 o o o o o v4           v3 --------- v4 
            int v1 = i*lev*numvtx_wd+j*lev; 
            int v2 = v1 + lev; 
            int v3 = v1 + lev*numvtx_wd; 
            int v4 = v3 + lev; 
            tri1[numtri1].v[0] = v1; 
            tri1[numtri1].v[1] = v2; 
            tri1[numtri1].v[2] = v3; 
            tri1[numtri1++].level = 0; 
            tri1[numtri1].v[0] = v4; 
            tri1[numtri1].v[1] = v2; 
            tri1[numtri1].v[2] = v3;        // we make sure that v[1]v[2] is always the longest edge in the triangle here 
            tri1[numtri1++].level = 0; 
        }
    int i = 0; // piin
    for (i=0; i<numimg; i++) 
    {
        imgused[i] = 0; 
        imglist[i] = i; 
    }

    // for vertices on the initial mesh, find nearby captured images
    for (i=0; i<numvtx_hc; i++) 
        for (int j=0; j<numvtx_wc; j++)
        {
            Vertex *pV = &vtx[i*lev*numvtx_wd+j*lev]; 
            find_nearby_cameras(pV, imglist, numimg); 
            for (int k=0; k<numinterp; k++) 
                imgused[pV->imgid[k]]=1;        // mark the related images as used (used in render_scene() later) 
        }
}

// project a vertex light ray to an image given its depth
// 'v' is the vertex, 'id' is the image id, 'dep' is the depth of the light ray, 'pos' is the resultant pixel location 
void proj_vtx2img(Vertex &v, int id, double dep, czsPt2 &pos)
{
    // use the depth of the light ray to determine the scene 3D position 
    double scale = (dep - rndpos.z)/v.dir.z; 
    czsPt3 wpos; 
    wpos.x = rndpos.x + scale*v.dir.x; 
    wpos.y = rndpos.y + scale*v.dir.y; 
    wpos.z = dep; 
    
    czsCalibPara *pcp = &cp[id]; 
    if (pcp->no_lens_distortion) 
    {   // if there is no lens distortion, we have a shotcut here using the combined projection matrix
        double *p = pcp->P.m; 
        double z = p[8]*wpos.x + p[9]*wpos.y + p[10]*wpos.z + p[11];
        if (z == 0.0) 
            exit_caview("Encounter divide by zero when projecting 3D points to images!\n"); 
        pos.x = (p[0]*wpos.x + p[1]*wpos.y + p[2]*wpos.z + p[3])/z; 
        pos.y = (p[4]*wpos.x + p[5]*wpos.y + p[6]*wpos.z + p[7])/z; 
        return; 
    }

    // the routine will enter here if there is lens distortion 
    // apply the external matrix first
    double *r = pcp->R.m; 
    double z = r[6]*wpos.x + r[7]*wpos.y + r[8]*wpos.z + pcp->T.z; 
    if (z == 0.0)
            exit_caview("Encounter divide by zero when projecting 3D points to images!\n"); 
    double x = (r[0]*wpos.x + r[1]*wpos.y + r[2]*wpos.z + pcp->T.x)/z; 
    double y = (r[3]*wpos.x + r[4]*wpos.y + r[5]*wpos.z + pcp->T.y)/z; 

    // geometry distortion 
    double r2 = x*x + y*y; 
    double r4 = r2*r2; 
    double r6 = r2*r4; 
    double xd = x*(1+pcp->kc[0]*r2+pcp->kc[1]*r4+pcp->kc[4]*r6) + 2*pcp->kc[2]*x*y+pcp->kc[3]*(r2+2*x*x); 
    double yd = y*(1+pcp->kc[0]*r2+pcp->kc[1]*r4+pcp->kc[4]*r6) + pcp->kc[2]*(r2+2*y*y)+2*pcp->kc[3]*x*y; 

    // apply the internal matrix last
    pos.x = pcp->fc[0]*(xd+pcp->alpha_c*yd) + pcp->cc[0]; 
    pos.y = pcp->fc[1]*yd + pcp->cc[1]; 
    return; 
}

// given a pixel location, get a patch around it 
// 'id' is the image id, 'pos' is the pixel location, 'pix' is an array to hold the returned data 
void getpatch(int id, czsPt2 &pos, double *pix)
{
    // we will round the pixel location to integers, such rounding is less accurate 
    // but runs at a faster rendering speed
    int x = int (pos.x + 0.5); 
    int y = int (pos.y + 0.5); 
    int xstart = x-(wndsize/2)*wndstep; 
    int ystart = y-(wndsize/2)*wndstep; 

    if (xstart < 0 || ystart < 0 || xstart+(wndsize-1)*wndstep > width || ystart+(wndsize-1)*wndstep > height) 
    {   // patch is outside the image 
        // we don't return partial data, instead, we simply set the patch to be invalid 
        pix[0] = -1.0; 
        return; 
    }

    unsigned char *data = &img[id].data[(ystart*width+xstart)*3]; 
    for (int i=0; i<wndsize; i++)   
    {   
        unsigned char *p = data + i*width*3*wndstep; 
        for (int j=0; j<wndsize; j++) 
        {
	//		*(pix++) =-0.169*p[0]-0.331*p[1]+0.5*p[2];  //use Cb 
      //    *(pix++) = 0.299*p[0]+0.587*p[1]+0.114*p[2];  //use Y value in YCrCb mode 
      *(pix++) = p[1];    // use green channel only to perform color consistency verification, as it's faster 
	//		 *(pix++) = max (p[0],p[1],p[2]); // use V in HSV
    //      *(pix++) = (p[0]+2*p[1]+p[2])/4;  //use average RGB 
            p += 3*wndstep; 
        }
    }   
}

// perform color consistency verification through sum of square errors 
double getccv_ssd(Vertex *pV, double *pix)
{
    int ww = wndsize*wndsize; 
    int count = 0; 
    double *p[MAX_NUM_INTERP]; 
    // how many patches are valid 
    for (int i=0; i<numinterp; i++) 
    {
        if (pix[i*ww] != -1.0) 
        {
            p[count] = &pix[i*ww]; 
            count ++; 
        }
    }
    if (count <= 1) return 1e5;     // return a big number if we cannot perform SSD

    double ssd = 0.0; 
	int i = 0;
    for (i=0; i<ww; i++) 
    {
        // find the mean of each pixel 
        double sum = 0.0; 
        for (int j=0; j<count; j++) 
            sum += p[j][i]; 
        double avg = sum/count; 
        double sum2 = 0.0; 
		int j = 0;
        for (j=0; j<count; j++) 
//            sum2 += fabs(p[j][i]-avg);              // absolute error 
            sum2 += (p[j][i]-avg)*(p[j][i]-avg);    // sqaure error 
        ssd += sum2/count; 
    }
    return (ssd/ww+1.0);    // minimum return value 1.0 
}

// perform color consistency verification through normalized mean-removed correlation 
double getccv_cor(Vertex *pV, double *pix)
{
    int ww = wndsize*wndsize; 
    int count = 0; 
    double *p[MAX_NUM_INTERP]; 
    // how many patches are valid 
    for (int i=0; i<numinterp; i++) 
    {
        if (pix[i*ww] != -1.0) 
        {
            p[count] = &pix[i*ww]; 
            count ++; 
        }
    }
    if (count <= 1) return 3.0;     // 3.0 is the largest value we will return 

    double avg[MAX_NUM_INTERP]; 
    double sum2[MAX_NUM_INTERP]; 
    double maxavg = -1, minavg=256; 
    // get the means of the pixel values
	int i = 0;
    for (i=0; i<count; i++) 
    {
        avg[i] = 0.0; 
        double *pp = p[i]; 
        for (int j=0; j<ww; j++)
            avg[i] += pp[j]; 
        avg[i] /= ww; 
        if (avg[i]<minavg) minavg = avg[i]; 
        if (avg[i]>maxavg) maxavg = avg[i]; 
    }
    if (maxavg - minavg > avgdiffth)
        return 3.0;     // early rejection 

    // remove means, calculate squared sum and detect uniform color regions 
    double max_sum2 = 0.0; 
    for (i=0; i<count; i++) 
    {
        sum2[i] = 0.0; 
        double *pp = p[i]; 
        for (int j=0; j<ww; j++) 
        {
            pp[j] -= avg[i]; 
            sum2[i] += pp[j]*pp[j]; 
        }
        if (sum2[i] > max_sum2) 
            max_sum2 = sum2[i]; 
    }
    if (max_sum2 < sum2th*ww) 
    {   // all the projections are uniform color, give a small constant return value for early acceptance 
        return 1.01;
    }

    double sumcorr = 0.0; 
	double tempp= 0.0;
    for (i=0; i<count; i++) 
    {
        if (sum2[i] == 0.0) continue;   // if this happens, we accurally assume a correlation value of 0.0 
        for (int j=i+1; j<count; j++)
        {
            if (sum2[j] == 0.0) continue; 
            double sumboth = 0.0; 
            double *pi = p[i], *pj = p[j];
            for (int k=0; k<ww; k++) 
               sumboth += pi[k]*pj[k]; 
		//	  sumboth += (pi[k]-avg[i])*(pj[k]-avg[j]); 
            sumcorr += sumboth/sqrt(sum2[i]*sum2[j]); 
        }
    }
	tempp=sumcorr/(count*(count-1)/2);
    return (2.0-sumcorr/(count*(count-1)/2));       // minimum return value is 1.0
 
}

// reconstruct the depth of a vertex 'pV' 
// 'mind' and 'maxd' are the minimum and maximum depth for searching the depth, which are entries for the 'depth' array 
void recon_vtxdepth(Vertex *pV, int mind, int maxd)
{
    int dlev = 1<<depthlevel; 
    int ww = wndsize*wndsize; 
    val[mind] = val[maxd+2] = 1e6;      // pad two big numbers at the beginning and end of the val array
    for (int d=mind; d<=maxd; d++)
    {
        for (int k=0; k<numinterp; k++) 
        {
            proj_vtx2img(*pV, pV->imgid[k], depth[d], projpos[k][d*dlev]); 
            getpatch(pV->imgid[k], projpos[k][d*dlev], &pixval[k*ww]); 
        }
        switch (ccv_mode)
        {
        case CCV_SSD: 
            val[d+1] = getccv_ssd(pV, pixval); 
            break; 
        case CCV_COR: 
            val[d+1] = getccv_cor(pV, pixval); 
            break; 
        }
    }
    int minsdep=0; 
    double minval1 = 1e6; 
	int d = 0;
    for (d=mind; d<=maxd; d++) 
        if (val[d+1] < minval1) 
        {
            minval1 = val[d+1]; 
            minsdep = d*dlev; 
        }
    double minval = minval1; 
    if (depthlevel != 0)
    {   // we will need to perform subdivision along the depth direction 
        // search for local minimum of val array 
        for (int d=mind; d<=maxd; d++) 
        {
            if (val[d] >= val[d+1] && val[d+1] <= val[d+2] && val[d+1] < 2*minval1)
            {   // it is a local minimum that worth subdivision, 
                // below we choose one of the depth intervals to subdivide
                int left, right; 
                double val_left, val_right; 
                if (val[d]<val[d+2]) 
                {   // choose depth layer d and d+1
                    left = (d-1)*dlev; 
                    right = d*dlev; 
                    val_left = val[d]; 
                    val_right = val[d+1]; 
                }
                else 
                {   // choose depth layer d+1 and d+2
                    left = d*dlev; 
                    right = (d+1)*dlev; 
                    val_left = val[d+1]; 
                    val_right = val[d+2]; 
                }
                // binarily and locally search the best depth
                int step = dlev/2; 
                while (step > 0)
                {
                    int mid = left + step; 
                    for (int k=0; k<numinterp; k++) 
                    {
                        proj_vtx2img(*pV, pV->imgid[k], sdepth[mid], projpos[k][mid]); 
                        getpatch(pV->imgid[k], projpos[k][mid], &pixval[k*ww]); 
                    }
                    double tmpval; 
                    switch (ccv_mode)
                    {
                    case CCV_SSD: 
                        tmpval = getccv_ssd(pV, pixval); 
                        break; 
                    case CCV_COR: 
                        tmpval = getccv_cor(pV, pixval); 
                        break; 
                    }
                    double tmpminval = CZ_MIN3(val_left, val_right, tmpval); 
                    if (val_left < val_right) 
                    {   // next will split the left half 
                        right = mid; 
                        val_right = tmpval; 
                    }
                    else
                    {   // next will split the right half 
                        left = mid; 
                        val_left = tmpval; 
                    }
                    if (tmpminval < minval) 
                    {
                        minval = tmpminval; 
                        if (minval == val_left) 
                            minsdep = left; 
                        else
                            minsdep = right; 
                    }
                    step /= 2; 
                }
            }
        }
    }
                
    pV->val = minval; 
    pV->depidx = minsdep; 
    // save the projection position for future usage (such as in rendering) 
    for (int k=0; k<numinterp; k++) 
        czfV2Set(pV->projpos[k], projpos[k][minsdep]); 
}

// calculate a weight based on where the pixel location 'pt' is 
// it will return 1 if the pixel is inside the image, and 0 if outside 
// at the boundary there is a smooth transition 
inline double weight_pixpos(czsPt2 &pt) 
{
    double xfrac = 2.0*fabs(pt.x-width/2)/width; 
    double yfrac = 2.0*fabs(pt.y-height/2)/height; 
    double xw = (xfrac < 0.9)? 1.0:((1-xfrac)*10.0); 
    xw = (xw<0)? 0:xw; 
    double yw = (yfrac < 0.9)? 1.0:((1-yfrac)*10.0); 
    yw = (yw<0)? 0:yw; 
    return (CZ_MIN2(xw,yw)); 
}

// test if a triangle needs to be subdivided 
// return 'true' if no need for subdivision, 'false' otherwise 
bool testtriangle(Triangle *pT) 
{
    if ((depth_mode == CONST_DEPTH) || (pT->level >= meshlevel))
        return true;    // no subdivision if performing constant depth rendering or the triangle is too small
    
    min_depidx = vtx[pT->v[0]].depidx; 
    max_depidx = min_depidx; 
    if (vtx[pT->v[1]].depidx > max_depidx) max_depidx = vtx[pT->v[1]].depidx; 
    if (vtx[pT->v[1]].depidx < min_depidx) min_depidx = vtx[pT->v[1]].depidx; 
    if (vtx[pT->v[2]].depidx > max_depidx) max_depidx = vtx[pT->v[2]].depidx; 
    if (vtx[pT->v[2]].depidx < min_depidx) min_depidx = vtx[pT->v[2]].depidx; 

    if (abs(max_depidx-min_depidx) > (1<<depthlevel)) 
        return false; 
    else
        return true; 

    // notice: min_depidx and max_depidx are global variables, they will be reused in subdivtriangle() 
    //         for determining the depth searching range of new vertices during the subdivision 
}

// subdivide a triangle of the mesh 
void subdivtriangle(Triangle *pT) 
{

    // find the reference images used by this triangle 
    // we will constrain that the new vertices only use these images as closeby images 
    int num = 0;        // num will store the number of reference images (un-duplicated)
    for (int i=0; i<3; i++) 
        for (int j=0; j<numinterp; j++) 
        {   
            int k = 0; 
            int id = vtx[pT->v[i]].imgid[j]; 
            while (k < num) 
            {
                if (imglist[k] == id) 
                    break;      // this images has been stored already 
                k++; 
            }
            if (k == num) 
            {
                imglist[num] = id; 
                num ++; 
            }
        }

    // the three new vertices during the subdivision (they are the mid-point of the edges) 
    int v[3]; 
    v[0] = (pT->v[1] + pT->v[2])/2; 
    v[1] = (pT->v[0] + pT->v[2])/2; 
    v[2] = (pT->v[0] + pT->v[1])/2; 
    // check if the three new vertices have already been processed, if not, process it
	int i = 0;
    for (i=0; i<3; i++) 
    {
        Vertex *pV = &vtx[v[i]]; 
        if (pV->val != 0.0) continue;      // has been processed before 
        find_nearby_cameras(pV, imglist, num); 
        // we make use of min_depidx and max_depidx obtained in testtriangle()
        int mind = min_depidx/(1<<depthlevel);      
        int maxd = max_depidx/(1<<depthlevel)+1; 
        if (maxd > numdepth-1) maxd = numdepth-1; 
        recon_vtxdepth(pV, mind, maxd); 
        
        // weight the vertex 
        double sum = 0.0; 
        for (int k=0; k<numinterp; k++) 
        {
            // pV->weight[k] was obtained during the call of find_nearby_cameras, 
            // here we reweight using pixel locations
            pV->weight[k] *= weight_pixpos(pV->projpos[k]); 
            sum += pV->weight[k]; 
        }
        if (sum == 0)
        {
            for (int k=0; k<numinterp; k++) 
                pV->weight[k] = 0.0; 
        }
        else 
        {   // normalize the weights
            for (int k=0; k<numinterp; k++) 
                pV->weight[k] /= sum; 
        }
    }

    // add the subdivided triangles to the tri1angles list 
    //                                          v1'
    //    v0 --------------- v2         v0 --------------- v2 
    //       |            /                |\    |      /
    //       |          /                  |  \  |    /
    //       |        /     subdivide      |    \|  /
    //       |      /      ==========>  v2'|----- /
    //       |    /                        |    /  v0'
    //       |  /                          |  /
    //    v1 |/                         v1 |/
    //
    tri1[numtri1].v[0] = v[2]; 
    tri1[numtri1].v[1] = pT->v[1]; 
    tri1[numtri1].v[2] = v[0]; 
    tri1[numtri1++].level = pT->level+1; 
    tri1[numtri1].v[0] = v[2]; 
    tri1[numtri1].v[1] = pT->v[0]; 
    tri1[numtri1].v[2] = v[0]; 
    tri1[numtri1++].level = pT->level+1; 
    tri1[numtri1].v[0] = v[1]; 
    tri1[numtri1].v[1] = pT->v[0]; 
    tri1[numtri1].v[2] = v[0]; 
    tri1[numtri1++].level = pT->level+1; 
    tri1[numtri1].v[0] = v[1]; 
    tri1[numtri1].v[1] = pT->v[2]; 
    tri1[numtri1].v[2] = v[0]; 
    tri1[numtri1++].level = pT->level+1; 
}

// process the triangles, subdivide them if necessary, or move them into the other triangle array 'tri2'
void processtriangles()
{
    int n = 0;
    while (n < numtri1)
    {
        Triangle *pT = &tri1[n];
        if (testtriangle(pT))
        {
            tri2[numtri2].v[0] = pT->v[0];
            tri2[numtri2].v[1] = pT->v[1]; 
            tri2[numtri2].v[2] = pT->v[2]; 
            tri2[numtri2++].level = pT->level;
        }
        else
		{
            subdivtriangle(pT); 
		}
        n++; 
    }
}

// reconstruct an adaptive mesh on the virtual imaging plane 
void recon_mesh()
{
    // reset the consistency scores of all the vertices to 0.0, which indicate the scores are invalid 
    int num = numvtx_wd*numvtx_hd; 
    for (int i=0; i<num; i++) 
        vtx[i].val = 0.0; 

    // obtain the depth of the hypothesis planes, notice the depth planes are also view-dependent 
    double mindist = mindepth - rndpos.z;   // it has been checked in displayfunc(), so mindist > 0 guaranteed
    double maxdist = maxdepth - rndpos.z;
    double invmindist = 1.0/mindist; 
    double invmaxdist = 1.0/maxdist; 
    double dstep_c = (invmindist - invmaxdist)/(numdepth-1); 
    int numdepth_d = (numdepth-1)*(1<<depthlevel)+1; 
    double dstep_d = (invmindist - invmaxdist)/(numdepth_d-1); 
    // the depths are equal distance in the disparity space 
	int i = 0;
    for (i=0; i<numdepth; i++) 
        depth[i] = 1.0/(invmaxdist + i*dstep_c) + rndpos.z;
    for (i=0; i<numdepth_d; i++) 
        sdepth[i] = 1.0/(invmaxdist + i*dstep_d) + rndpos.z; 
    



    int lev = 1<<meshlevel; 
    if (depth_mode == CONST_DEPTH)
    {   // constant depth rendering 
        for (int i=0; i<numvtx_hc; i++) 
            for (int j=0; j<numvtx_wc; j++) 
            {
                Vertex *pV = &vtx[i*lev*numvtx_wd+j*lev]; 
                double sum = 0.0; 
                for (int k=0; k<numinterp; k++) 
                {
                    proj_vtx2img(*pV, pV->imgid[k], constdepth, pV->projpos[k]); 
                    pV->weight[k] *= weight_pixpos(pV->projpos[k]); 
                    sum += pV->weight[k]; 
                }
                if (sum == 0) 
                {
                    for (int k=0; k<numinterp; k++) 
                        pV->weight[k] = 0.0; 
                }
                else 
                {   // normalize the weights 
                    for (int k=0; k<numinterp; k++) 
                        pV->weight[k] /= sum; 
                }
                pV->val = 1.0;      // some arbitrary number 
            }
    }
    else
    {
        // obtain the depths of the vertices on the initial mesh 
        for (int i=0; i<numvtx_hc; i++) 
            for (int j=0; j<numvtx_wc; j++) 
            {
                Vertex *pV = &vtx[i*lev*numvtx_wd+j*lev]; 
                recon_vtxdepth(pV, 0, numdepth-1);  // full range search for these vertices 

                double sum = 0.0; 
                for (int k=0; k<numinterp; k++) 
                {
                    pV->weight[k] *= weight_pixpos(pV->projpos[k]); 
                    sum += pV->weight[k]; 
                }
                if (sum == 0)
                {
                    for (int k=0; k<numinterp; k++) 
                        pV->weight[k] = 0.0; 
                }
                else 
                {   // normalize the weights
                    for (int k=0; k<numinterp; k++) 
                        pV->weight[k] /= sum; 
                }
            }
    }

    processtriangles(); 
}

// used in complete_mesh(), if for a triangle, one of its edges is split 
void one_edge_proc(int v0, int v1, int v2, int l)   // we assume the edge v0v1 has split 
{
    int lev = 1<<(meshlevel-l); 
    int vtmp1 = v0, vtmp2; 
    int step = (v1-v0)/lev; 
    for (int i=1; i<=lev; i++)
    {
        vtmp2 = v0+i*step; 
        if(vtx[vtmp2].val!=0.0)
        {
            tri1[numtri1].v[0] = v2; 
            tri1[numtri1].v[1] = vtmp1; 
            tri1[numtri1++].v[2] = vtmp2; 
            vtmp1 = vtmp2; 
        }
    }
}

// used in complete_mesh(), if for a triangle, two of its edges are split 
void two_edge_proc(int v0, int v1, int v2, int l)   
// we assume the edges v0v1 and v0v2 have split
{
    int lev = 1<<(meshlevel-l); 
    int v01 = (v0+v1)/2; 
    int v02 = (v0+v2)/2; 
    one_edge_proc(v01, v1, v2, l+1); 
    one_edge_proc(v02, v2, v01, l+1); 
    tri2[numtri2].v[0] = v0; 
    tri2[numtri2].v[1] = v01; 
    tri2[numtri2].v[2] = v02; 
    tri2[numtri2++].level = l+1; 
}

// used in complete_mesh(), if for a triangle, all its edges are split 
void three_edge_proc(int v0, int v1, int v2, int l)
{
    int v01 = (v0+v1)/2; 
    int v02 = (v0+v2)/2; 
    int v12 = (v1+v2)/2; 
    one_edge_proc(v0, v01, v12, l+1); 
    one_edge_proc(v0, v02, v12, l+1); 
    tri2[numtri2].v[0] = v1; 
    tri2[numtri2].v[1] = v01; 
    tri2[numtri2].v[2] = v12; 
    tri2[numtri2++].level = l+1; 
    tri2[numtri2].v[0] = v2; 
    tri2[numtri2].v[1] = v02; 
    tri2[numtri2].v[2] = v12; 
    tri2[numtri2++].level = l+1; 
}

// although we may use the triangles in 'tri2' array to render directly, this mesh 
// may has some T-intersections, which may cause cracks in the rendered image
// complete_mesh() is used to remove these T-intersections by inserting more triangles (but no additional vertices) 
// after this function, all the to-be-rendered triangles are all in the 'tri1' array 
void complete_mesh()
{
    numtri1 = 0; 
    int n = 0; 
    while (n < numtri2) 
    {
        Triangle *pT = &tri2[n]; 
        int lev = 1<<(meshlevel-pT->level); 
        if (lev == 1) 
        {   // the triangle is the smallest, no further processing, insert it to 'tri1' for final rendering 
            tri1[numtri1].v[0] = pT->v[0]; 
            tri1[numtri1].v[1] = pT->v[1]; 
            tri1[numtri1++].v[2] = pT->v[2]; 
            n++; 
            continue; 
        }
        // get midpoints of the edges 
        int v[3]; 
        v[0] = (pT->v[1] + pT->v[2])/2; 
        v[1] = (pT->v[0] + pT->v[2])/2; 
        v[2] = (pT->v[0] + pT->v[1])/2; 
        int tmp = ((vtx[v[2]].val!=0.0)*4)+((vtx[v[1]].val!=0.0)*2)+(vtx[v[0]].val!=0.0); 
        switch (tmp) 
        {
        case 0:         // all midpoints are not processed, the triangle needs no further processing
            tri1[numtri1].v[0] = pT->v[0]; 
            tri1[numtri1].v[1] = pT->v[1]; 
            tri1[numtri1++].v[2] = pT->v[2]; 
            break; 
        case 1:         // v[0] has been processed, but v[1] and v[2] have not 
            one_edge_proc(pT->v[1], pT->v[2], pT->v[0], pT->level); 
            break; 
        case 2:         // v[1] has been processed, but v[0] and v[2] have not 
            one_edge_proc(pT->v[2], pT->v[0], pT->v[1], pT->level); 
            break;      
        case 4:         // v[2] has been processed, but v[0] and v[1] have not 
            one_edge_proc(pT->v[0], pT->v[1], pT->v[2], pT->level); 
            break;      
        case 3:         // v[0] and v[1] have been processed, but v[2] has not 
            two_edge_proc(pT->v[2], pT->v[1], pT->v[0], pT->level); 
            break; 
        case 5:         // v[0] and v[2] have been processed, but v[1] has not 
            two_edge_proc(pT->v[1], pT->v[2], pT->v[0], pT->level); 
            break; 
        case 6:         // v[1] and v[2] have been processed, but v[0] has not 
            two_edge_proc(pT->v[0], pT->v[1], pT->v[2], pT->level);  
            break; 
        case 7:         // all the midpoint vertices were processed 
            three_edge_proc(pT->v[0], pT->v[1], pT->v[2], pT->level); 
            break; 
        }
        n++; 
    }
}

// render the scene 
void render_scene()
{
    glClear (GL_COLOR_BUFFER_BIT); 

    if (rnd_mode & RND_SCENE) 
    {
		glEnable (GL_BLEND);
		glBlendFunc (GL_SRC_ALPHA, GL_ONE);
        glEnable (GL_TEXTURE_2D); 
		

        // use the textures one by one to render the scene
        for (int i=0; i<numimg; i++) 
        {
            if (!imgused[i]) 
				continue; 
            
            glBindTexture(GL_TEXTURE_2D, texname[i]); 
		

            glBegin(GL_TRIANGLES); 
            for (int j=0; j<numtri1; j++) 
            {
                // check if the triangle needs this texture to render, if yes, what are the weights of the three vertices
                int used = 0; 
                double w[3]={0.0,0.0,0.0};
                int proj[3]={0,0,0}; 
                czsPt2 pos[3]; 
                for (int k=0; k<3; k++) 
                {
                    Vertex *pV = &vtx[tri1[j].v[k]]; 
                    for (int n=0; n<numinterp; n++) 
                    {
                        if (pV->imgid[n] == i) 
                        {
                            w[k] = pV->weight[n]; 
                            czfV2Set(pos[k], pV->projpos[n]); 
                            used = 1; 
                            proj[k] = 1; 
                            break; 
                        }
                    }
                }
                if (used) 
                {
                    for (int k=0; k<3; k++) 
                    {
                        Vertex *pV = &vtx[tri1[j].v[k]]; 
                        if (!proj[k])
                        {   // the vertex hasnot been projected to this image before 
                            if (depth_mode == CONST_DEPTH) 
                                proj_vtx2img(*pV, i, constdepth, pos[k]);
                            else 
                                proj_vtx2img(*pV, i, sdepth[pV->depidx], pos[k]); 
                        }
                        GLdouble x = (pos[k].x-width/2)/texwidth+0.5; 
                        GLdouble y = (pos[k].y-height/2)/texheight+0.5; 
                        glColor4d (1.0, 1.0, 1.0, w[k]); 
                        glTexCoord2d (x, y);
						glVertex3f (pV->pos.x, -pV->pos.y, -1.0); 

                    }
                }   
            }
            glEnd(); 

        }
		glDisable (GL_TEXTURE_2D); 
        glDisable (GL_BLEND); 
    }

    if (rnd_mode & RND_FRAME)      // overlay the tiangulized frame 
    {
        double invmindist = 1.0/(mindepth-rndpos.z); 
        double invmaxdist = 1.0/(maxdepth-rndpos.z); 
        
        for (int i=0; i<numtri1; i++) 
        {
            glBegin(GL_LINE_LOOP); 
            for (int j=0; j<3; j++) 
            {
                Vertex *pV = &vtx[tri1[i].v[j]]; 
                double invd = 1.0/(sdepth[pV->depidx]-rndpos.z); 
                double color = (invd-invmaxdist)/(invmindist-invmaxdist);  
                glColor3d (color, color, color); 
                glVertex3d (pV->pos.x, -pV->pos.y, -1.0); 
            }
            glEnd(); 
        }
    }

	

    glutSwapBuffers(); 
}

// display the scene 
void displayfunc () 
{
    if (depth_mode == CONST_DEPTH && rndpos.z >= constdepth)
    {
        printf ("The virtual viewpoint has been beyond the constant rendering depth! Please backward or change the rendering depth!\n"); 
        // show nothing 
        glClear (GL_COLOR_BUFFER_BIT); 
        glutSwapBuffers(); 
    }
    else if (depth_mode == RECON_DEPTH && rndpos.z >= mindepth) 
    {   // should never enter the scene!
        printf ("The virtual viewpoint has been beyond the minimum scene depth! Please backward!\n"); 
        // show nothing
        glClear (GL_COLOR_BUFFER_BIT); 
        glutSwapBuffers(); 
    }
    else 
    {
        // construct the initial mesh 
        construct_initmesh();
		
        // reconstruct the mesh geometry 
        recon_mesh(); 

        // complete the mesh geometry, need to remove the T-intersections exhibited in the mesh we obtained abov
        complete_mesh(); 

        // render the scene 
        render_scene(); 

    }
}
int gGlutWnd;

void change_depth(int id)
{
	if(depth_mode != RECON_DEPTH)
	{
//		gGlutWnd = window;
//		glutSetWindow(gGlutWnd);
		depth_change = gui_depth - old_depth_increase;
		constdepth += depth_change*stepfwd;
		printf("gui_depth = %d, old_depth = %d, depth_chage = %d\n",gui_depth, old_depth_increase, depth_change);
		old_depth_increase = gui_depth;	
	}
	glutPostRedisplay();
}

void constant_depth(int id)
{
	if(gui_depth_mode == 1)
		depth_mode = CONST_DEPTH;
	if(gui_depth_mode == 0)
		depth_mode = RECON_DEPTH;
	glutPostRedisplay();
}

void change_view(int id)
{
	if(gui_view == 0)
		rnd_mode = RND_SCENE; 
	if(gui_view == 1)
		rnd_mode = RND_FRAME; 
	if(gui_view == 2)
		rnd_mode = RND_BOTH;

	glutPostRedisplay();
}
void reset_scene()
{
	czfQuatCamReset(rqcam); 
    constdepth = dftdepth; 
    // locate the rendering position
    // notice rqcam is in OpenGL coordinate, rndpos is in our world coordinate 
    rndpos.x = rqcam.pos.x + allcam_ctr.x; 
    rndpos.y = -rqcam.pos.y + allcam_ctr.y; 
    rndpos.z = -rqcam.pos.z + allcam_ctr.z; 
	glutPostRedisplay();
}
void translation_z(int id)
{
	int diff_z =(int) (matrixT[0]-old_matrixT[0]);
	printf("%d",diff_z);
	czfQuatCamTranslate(rqcam, 0, 0, -stepside*diff_z);
	rndpos.z = -rqcam.pos.z + allcam_ctr.z;
	old_matrixT[0] = matrixT[0];

	glutPostRedisplay();
}

	
int main(int argc, char *argv[]) 
{
    if (argc != 2) 
    {
        printf ("Usage: %s configurefile\n", argv[0]); 
        exit (0); 
    }
    // load the configuration file and related image files 
    if (!loadscene (argv[1]))
    {
        printf ("Failed to load the scene with configuration file %s!\n", argv[1]); 
        exit_caview(); 
    }
    
    glutInit(&argc, argv); 
    glutInitDisplayMode (GLUT_DOUBLE | GLUT_RGB); 
    glutInitWindowSize (width*2, height*2); 
    glutInitWindowPosition(100,100); 
    window = glutCreateWindow("AMP CAView"); 

    

	GLUI *glui = GLUI_Master.create_glui( "GLUI", 0, 105+ width*2, 100);
	glui->set_main_gfx_window(window);
	glui->add_statictext("Caview GUI");
	glui->add_separator();
//	glui->add_rotation("Rotation", matrixR);
	glui->add_translation("Translation Foward/Backward", GLUI_TRANSLATION_Y, matrixT, 1, translation_z);
	
	GLUI_Rollout *rollout = glui->add_rollout("Constant Depth Functions", false,1 );
	glui->add_checkbox_to_panel(rollout,"Enable Constant Depth",&gui_depth_mode, 1,constant_depth);
	GLUI_Spinner * depth_spinner = glui->add_spinner_to_panel(rollout, "Depth", GLUI_SPINNER_INT, &gui_depth,1,change_depth);
	depth_spinner->set_int_limits(-100, 100, GLUI_LIMIT_CLAMP );
	depth_spinner->set_speed(0.1);

	GLUI_RadioGroup *radiogroup = glui->add_radiogroup(&gui_view,1,change_view);
	glui->add_radiobutton_to_group(radiogroup,"Rendered View");
	glui->add_radiobutton_to_group(radiogroup,"Wireframe View");
	glui->add_radiobutton_to_group(radiogroup,"Both");

	glui->add_button("Reset Scene",1,(GLUI_Update_CB)reset_scene);  

    glutKeyboardFunc(keyboardfunc); 
    glutSpecialFunc(specialfunc);     
    glutDisplayFunc(displayfunc); 
	glutMotionFunc(motionFunc);
    glutMouseFunc(mouseFunc);

	init_rendering(); 

    glutMainLoop(); 

	return 0;
}
