/********************************************************************************
 * Copyright (c) 2004, Cha Zhang <czhang@ieee.org>, Advanced Multimedia 
 * Processing Lab, ECE, Carnegie Mellon University. 
 * All rights reserved. 
 * 
 * History
 * -------
 *
 * 2004.5.15 Created by Cha Zhang at CMU as a simplified version of the
 *           rendering algorithm implemented for the mobile camera array
 *           at AMP Lab, ECE, Carnegie Mellon University. 
 ********************************************************************************/

/********************************************************************************
 * This is the cpp file for quaternion camera. 
 ********************************************************************************/

#ifdef _WIN32
#include "windows.h" 
#endif

#include <GL/gl.h>
#include <GL/glu.h>
#include "cz_quatcam_utils.h"

// this quaternion camera follows OpenGL coordinate system 

void czfQuatCamGet(const czsQuatCam &qcam, czsQuat &q)
{
    double c = cos (0.5*qcam.rot_angle); 
    double s = sin (0.5*qcam.rot_angle); 
	double val = s / czfV3Len(qcam.rot_axis); 

	q.w = c;
	q.x = qcam.rot_axis.x * val;
	q.y = qcam.rot_axis.y * val;
	q.z = qcam.rot_axis.z * val; 
}

void czfQuatCamSet(czsQuatCam &qcam, const czsQuat &q)
{
    qcam.rot_angle = 2.0 * acos(q.w);
	if (qcam.rot_angle == 0.0) 
    {
        czfV3Set (qcam.rot_axis, 0.0, 0.0, -1.0); 
        return; 
    }    

    qcam.rot_axis.x = q.x; 
    qcam.rot_axis.y = q.y; 
    qcam.rot_axis.z = q.z; 
    czfV3Norm(qcam.rot_axis); 
}

void czfQuatCamSetInit(czsQuatCam &qcam, const czsPt3 &pos, const double rotangle, const czsVec3 &rotaxis, const double fov)
{
    czfV3Set (qcam.init_pos, pos); 
    qcam.init_rot_angle = rotangle; 
    czfV3Set (qcam.init_rot_axis, rotaxis); 
    qcam.init_fov = fov; 
}

void czfQuatCamReset(czsQuatCam &qcam)
{
    czfV3Set (qcam.pos, qcam.init_pos); 
    qcam.rot_angle = qcam.init_rot_angle; 
    czfV3Set (qcam.rot_axis, qcam.init_rot_axis); 
    qcam.fov = qcam.init_fov; 
}

void czfQuatCamSetGLMat(const czsQuatCam &qcam, const double znear, const double zfar, const double width, const double height)
{
	glMatrixMode(GL_PROJECTION);
    glLoadIdentity(); 
    gluPerspective (CZ_R2D*qcam.fov, width/height, znear, zfar); 

	glMatrixMode(GL_MODELVIEW);
    glLoadIdentity(); 
	glRotated(-CZ_R2D*qcam.rot_angle, qcam.rot_axis.x, qcam.rot_axis.y, qcam.rot_axis.z);
	glTranslated(-qcam.pos.x, -qcam.pos.y, -qcam.pos.z);
}

void czfQuatCamGetOrthoDirs(const czsQuatCam &qcam, czsVec3 &viewdir, czsVec3 &updir, czsVec3 &rightdir)
{
	czsQuat q;
	czfQuatCamGet (qcam, q);

    viewdir.x = viewdir.y = 0; viewdir.z = -1.0; 
    czfQuatRot (viewdir, q); 

    updir.x = updir.z = 0; updir.y = 1.0;    
    czfQuatRot (updir, q); 

	czfV3Cross(viewdir, updir, rightdir);
}

void czfQuatCamTranslate(czsQuatCam &qcam, const double dx, const double dy, const double dz)
{
    // dx, dy, dz is in OpenGL coordinate, so the direction of dz is different from viewdir, which reverses zdir
    czsVec3 viewdir, updir, rightdir; 
    czfQuatCamGetOrthoDirs(qcam, viewdir, updir, rightdir); 
 	qcam.pos.x += rightdir.x*dx + updir.x*dy - viewdir.x*dz;
 	qcam.pos.y += rightdir.y*dx + updir.y*dy - viewdir.y*dz;
 	qcam.pos.z += rightdir.z*dx + updir.z*dy - viewdir.z*dz;
}

void czfQuatCamRotate(czsQuatCam &qcam, const czsQuat &q, const double rotdist)
{
	czsVec3 viewdir, updir, rightdir;
	czfQuatCamGetOrthoDirs(qcam, viewdir, updir, rightdir);

	czsPt3 rotcenter; 
    czfV3Set (rotcenter, qcam.pos.x+rotdist*viewdir.x, qcam.pos.y+rotdist*viewdir.y, qcam.pos.z+rotdist*viewdir.z); 

	czsQuat currq, rotq;
	czfQuatCamGet (qcam, currq); 
	czfQuatComp(q, currq, rotq);
	czfQuatCamSet (qcam, rotq);

	czfQuatCamGetOrthoDirs(qcam, viewdir, updir, rightdir);

	qcam.pos.x = rotcenter.x - rotdist*viewdir.x;
	qcam.pos.y = rotcenter.y - rotdist*viewdir.y;
    qcam.pos.z = rotcenter.z - rotdist*viewdir.z;
}

void czfQuatCamZoom(czsQuatCam &qcam, const double logzoom)
{
    qcam.fov *= exp(logzoom); 
}

